from functools import lru_cache

JID_ESCAPE_SEQUENCES = {
    "\\20",
    "\\22",
    "\\26",
    "\\27",
    "\\2f",
    "\\3a",
    "\\3c",
    "\\3e",
    "\\40",
    "\\5c",
}


JID_UNESCAPE_TRANSFORMATIONS = {
    "\\20": " ",
    "\\22": '"',
    "\\26": "&",
    "\\27": "'",
    "\\2f": "/",
    "\\3a": ":",
    "\\3c": "<",
    "\\3e": ">",
    "\\40": "@",
    "\\5c": "\\",
}


@lru_cache(1000)
def unescape_node(node: str) -> str:
    """Unescape a local portion of a JID."""
    unescaped = []
    seq = ""
    for i, char in enumerate(node):
        if char == "\\":
            seq = node[i : i + 3]
            if seq not in JID_ESCAPE_SEQUENCES:
                seq = ""
        if seq:
            if len(seq) == 3:
                unescaped.append(JID_UNESCAPE_TRANSFORMATIONS.get(seq, char))

            # Pop character off the escape sequence, and ignore it
            seq = seq[1:]
        else:
            unescaped.append(char)
    return "".join(unescaped)


ESCAPE_TABLE = "".maketrans({v: k for k, v in JID_UNESCAPE_TRANSFORMATIONS.items()})



import pytest
from slixmpp.exceptions import XMPPError

from conftest import AvatarFixtureMixin, UUIDFixtureMixin
from slixmpp import JID, Iq

from slidge import BaseGateway, BaseSession, GatewayUser
from slidge.contact import LegacyContact
from slidge.group import LegacyBookmarks, LegacyMUC
from slidge.util.test import SlidgeTest


class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"
    GROUPS = True


class Session(BaseSession):
    async def login(self):
        return "YUP"


class Contact(LegacyContact):
    async def update_info(self):
        if self.legacy_id.startswith("room"):
            raise XMPPError
        if self.legacy_id == "no-nick":
            self.online()
            return
        self.name = "duplicate"


class MUC(LegacyMUC):
    _ALL_INFO_FILLED_ON_STARTUP = True

    async def update_participants(self):
        if self.legacy_id == "room-occupant-id":
            await self.get_participant("nick", occupant_id="unique-id-1")
            await self.get_participant("nick", occupant_id="unique-id-2")
        elif self.legacy_id == "room-user-nick-taken":
            await self.get_participant("user", occupant_id="unique-id-1")
            await self.get_user_participant(occupant_id="user-id")
        else:
            await self.get_participant_by_legacy_id("dup1")
            dup = await self.get_participant(occupant_id="unique-id")
            dup.nickname = "duplicate"


class Bookmarks(LegacyBookmarks):
    ROOMS = ["room", "room-occupant-id", "room-user-nick-taken"]

    async def fill(self):
        for legacy_id in Bookmarks.ROOMS:
            muc = await self.by_legacy_id(legacy_id)
            await muc.update_participants()


@pytest.mark.usefixtures("avatar")
class TestMUCAllInfoOnStartup(UUIDFixtureMixin, AvatarFixtureMixin, SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("user@test").bare,
                legacy_module_data={"username": "myname", "city": ""},
                preferences={"sync_avatar": False, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()
        self.run_coro(
            self.xmpp._BaseGateway__dispatcher._on_user_register(
                Iq(sfrom="user@test")
            )
        )
        welcome = self.next_sent()
        assert welcome["body"]
        stanza = self.next_sent()
        assert "logging in" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing contacts" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing groups" in stanza["status"].lower(), stanza
        probe = self.next_sent()
        assert probe.get_type() == "probe"
        stanza = self.next_sent()
        assert "yup" in stanza["status"].lower(), stanza


    @property
    def user_session(self) -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("user@test")
        )

    def test_join(self):
        self.recv(  # language=XML
            f"""
            <presence from="{self.user_session.user_jid}/client"
                      to="room@aim.shakespeare.lit/mynick">
              <x xmlns='http://jabber.org/protocol/muc' />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <presence from="room@aim.shakespeare.lit/duplicate"
                      to="user@test/client">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="dup1@aim.shakespeare.lit/slidge" />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <presence from="room@aim.shakespeare.lit/duplicate (unique-id)"
                      to="user@test/client">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="unique-id" />
            </presence>
            """
        )
        self.send(  # language=XML
            f"""
            <presence from="room@aim.shakespeare.lit/user"
                      to="user@test/client">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant" />
                <status code="210" />
                <status code="110" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="slidge-user" />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <message type="groupchat"
                     from="room@aim.shakespeare.lit"
                     to="user@test/client">
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="5"
                         by="room@aim.shakespeare.lit" />
              <subject />
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="room" />
            </message>
            """,
            use_values=False,
        )
        self.send("""
            <presence xmlns="jabber:component:accept"
                      to="user@test/client"
                      from="room@aim.shakespeare.lit">
              <x xmlns="vcard-temp:x:update">
                <photo />
              </x>
            </presence>
            """, use_values=False)
        self.recv(  # language=XML
            """
            <iq type="get"
                from="user@test/client"
                to="room@aim.shakespeare.lit"
                id="123">
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="room@aim.shakespeare.lit"
                to="user@test/client"
                id="123">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="conference"
                          type="text" />
                <feature var="http://jabber.org/protocol/muc" />
                <feature var="http://jabber.org/protocol/muc#stable_id" />
                <feature var="http://jabber.org/protocol/muc#self-ping-optimization" />
                <feature var="urn:xmpp:mam:2" />
                <feature var="urn:xmpp:mam:2#extended" />
                <feature var="urn:xmpp:sid:0" />
                <feature var="muc_persistent" />
                <feature var="vcard-temp" />
                <feature var="urn:xmpp:ping" />
                <feature var="urn:xmpp:occupant-id:0" />
                <feature var="jabber:iq:register" />
                <feature var="urn:xmpp:message-moderate:1" />
                <feature var="muc_open" />
                <feature var="muc_semianonymous" />
                <feature var="muc_public" />
                <x xmlns="jabber:x:data"
                   type="result">
                  <field var="FORM_TYPE"
                         type="hidden">
                    <value>http://jabber.org/protocol/muc#roominfo</value>
                  </field>
                  <field var="muc#roomconfig_persistentroom"
                         type="boolean">
                    <value>1</value>
                  </field>
                  <field var="muc#roomconfig_changesubject"
                         type="boolean">
                    <value>0</value>
                  </field>
                  <field var="muc#maxhistoryfetch">
                    <value>100</value>
                  </field>
                  <field var="muc#roominfo_subjectmod"
                         type="boolean">
                    <value>0</value>
                  </field>
                  <field var="muc#roominfo_occupants">
                    <value>3</value>
                  </field>
                  <field var="muc#roomconfig_membersonly"
                         type="boolean">
                    <value>0</value>
                  </field>
                  <field var="muc#roomconfig_whois"
                         type="list-single">
                    <value>moderators</value>
                  </field>
                  <field var="muc#roomconfig_publicroom"
                         type="boolean">
                    <value>1</value>
                  </field>
                  <field var="muc#roomconfig_allowpm"
                         type="boolean">
                    <value>0</value>
                  </field>
                </x>
              </query>
            </iq>
            """
        )

    def test_join_occupant_id(self):
        self.recv(  # language=XML
            f"""
            <presence from="{self.user_session.user_jid}/client"
                      to="room-occupant-id@aim.shakespeare.lit/mynick">
              <x xmlns='http://jabber.org/protocol/muc' />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <presence from="room-occupant-id@aim.shakespeare.lit/nick"
                      to="user@test/client">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="unique-id-1" />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <presence from="room-occupant-id@aim.shakespeare.lit/nick (unique-id-2)"
                      to="user@test/client">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="unique-id-2" />
            </presence>
            """
        )
        self.send(  # language=XML
            f"""
            <presence from="room-occupant-id@aim.shakespeare.lit/user"
                      to="user@test/client">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant" />
                <status code="210" />
                <status code="110" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="slidge-user" />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <message type="groupchat"
                     from="room-occupant-id@aim.shakespeare.lit"
                     to="user@test/client">
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="5"
                         by="room-occupant-id@aim.shakespeare.lit" />
              <subject />
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="room" />
            </message>
            """,
            use_values=False,
        )

    def test_updating_info(self):
        contact = self.run_coro(self.user_session.contacts.by_legacy_id("no-nick"))
        with contact.updating_info():
            contact.name = "prout"
        assert contact.name == "prout"
        contact = self.run_coro(self.user_session.contacts.by_legacy_id("no-nick"))
        assert contact.name == "prout"


class TestMUCUserNickIsTaken(UUIDFixtureMixin, AvatarFixtureMixin, SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("user@test").bare,
                legacy_module_data={"username": "myname", "city": ""},
                preferences={"sync_avatar": False, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()
        self.run_coro(
            self.xmpp._BaseGateway__dispatcher._on_user_register(
                Iq(sfrom="user@test")
            )
        )
        welcome = self.next_sent()
        assert welcome["body"]
        stanza = self.next_sent()
        assert "logging in" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing contacts" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing groups" in stanza["status"].lower(), stanza
        probe = self.next_sent()
        assert probe.get_type() == "probe"
        stanza = self.next_sent()
        assert "yup" in stanza["status"].lower(), stanza

    def test_user_nick_is_taken(self):
        self.recv(  # language=XML
            f"""
            <presence from="user@test/client"
                      to="room-user-nick-taken@aim.shakespeare.lit/user">
              <x xmlns='http://jabber.org/protocol/muc' />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <presence from="room-user-nick-taken@aim.shakespeare.lit/user"
                      to="user@test/client">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="unique-id-1" />
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <presence from="room-user-nick-taken@aim.shakespeare.lit/user (user-id)"
                      to="user@test/client">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant" />
                <status code="110" />
                <status code="210" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="user-id" />
            </presence>
            """,
            use_values=False,
        )
        assert self.next_sent() is not None # subject
        assert self.next_sent() is not None # room avatar



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from enum import Enum, IntEnum
from pathlib import Path

from slixmpp import ComponentXMPP

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

# sys.path.insert(0, os.path.abspath('.'))
# sys.path.insert(0, os.path.abspath('..'))
# sys.path.insert(0, os.path.abspath('../..'))


# -- Project information -----------------------------------------------------

project = "Slidge"
copyright = "2025, the slidge contributors"
author = "Nicolas Cedilnik"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    # "sphinx.ext.viewcode",  # crashes build unfortunately
    "sphinx.ext.autodoc.typehints",
    "sphinxarg.ext",
    "autoapi.extension",
    "slidge_sphinx_extensions.config_obj",
]

autodoc_typehints = "description"

# Incldude __init__ docstrings
# autoclass_content = "class"
# autoapi_python_class_content = "class"

autoapi_type = "python"
autoapi_dirs = ["../../slidge", "../../superduper"]
autoapi_add_toctree_entry = False
autoapi_keep_files = False
autoapi_root = "api/auto"
autoapi_ignore = ["*xep_*", "slidge/core/*"]
autoapi_options = [
    "members",
    "show-module-summary",
    "inherited-members",
    "imported-members",
    # these on-by-default parameters are disabled
    # "undoc-members",
    # "private-members",
    # "show-inheritance",
    # "special-members",
]


def skip_stuff(app, what, name, obj, skip, options):
    # Hide some stuff from the docs
    if name.endswith(".Register"):
        skip = True
    elif name.endswith("user_store"):
        skip = True
    elif any(name.endswith("Gateway." + x) for x in dir(ComponentXMPP)):
        skip = True
    elif any(
        name.endswith("MucType." + x)
        or name.endswith("CommandAccess." + x)
        or name.endswith("RegistrationType." + x)
        for x in dir(int) + dir(Enum) + dir(IntEnum) + ["name", "value"]
    ):
        skip = True
    return skip


def setup(sphinx):
    sphinx.connect("autoapi-skip-member", skip_stuff)


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "slixmpp": ("https://slixmpp.readthedocs.io/en/latest/", None),
}

extlinks = {"xep": ("https://xmpp.org/extensions/xep-%s.html", "XEP-%s")}
html_theme = "furo"
html_theme_options = {
    "source_edit_link": "https://codeberg.org/slidge/slidge/_edit/main/docs/source/{filename}",
    "source_view_link": "https://codeberg.org/slidge/slidge/src/branch/main/docs/source/{filename}",
    "footer_icons": [
        {
            "name": "Codeberg",
            "url": "https://codeberg.org/slidge/slidge",
            "html": Path("codeberg.svg").read_text(),
        },
    ],
}



# Commands available to users
from copy import deepcopy
from typing import TYPE_CHECKING, Any, cast

from slixmpp import JID
from slixmpp.exceptions import XMPPError

from ..group.room import LegacyMUC
from ..util.types import AnyBaseSession, LegacyGroupIdType, MucType, UserPreferences
from .base import (
    Command,
    CommandAccess,
    Confirmation,
    Form,
    FormField,
    FormValues,
    SearchResult,
    TableResult,
)
from .categories import CONTACTS, GROUPS

if TYPE_CHECKING:
    pass


class Search(Command):
    NAME = "🔎 Search for contacts"
    HELP = "Search for contacts via this gateway"
    CHAT_COMMAND = "find"
    NODE = CONTACTS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = CONTACTS

    async def run(
        self, session: AnyBaseSession | None, _ifrom: JID, *args: str
    ) -> Form | SearchResult | None:
        assert session is not None
        await session.ready
        if args:
            return await session.on_search(
                {self.xmpp.SEARCH_FIELDS[0].var: " ".join(args)}
            )
        return Form(
            title=self.xmpp.SEARCH_TITLE,
            instructions=self.xmpp.SEARCH_INSTRUCTIONS,
            fields=self.xmpp.SEARCH_FIELDS,
            handler=self.search,
        )

    @staticmethod
    async def search(
        form_values: FormValues, session: AnyBaseSession | None, _ifrom: JID
    ) -> SearchResult:
        assert session is not None
        results = await session.on_search(form_values)  # type: ignore
        if results is None:
            raise XMPPError("item-not-found", "No contact was found")

        return results


class SyncContacts(Command):
    NAME = "🔄 Sync XMPP roster"
    HELP = (
        "Synchronize your XMPP roster with your legacy contacts. "
        "Slidge will only add/remove/modify contacts in its dedicated roster group"
    )
    CHAT_COMMAND = "sync-contacts"
    NODE = CONTACTS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = CONTACTS

    async def run(self, session: AnyBaseSession | None, _ifrom, *_) -> Confirmation:
        assert session is not None
        await session.ready
        return Confirmation(
            prompt="Are you sure you want to sync your roster?",
            success=None,
            handler=self.sync,
        )

    async def sync(self, session: AnyBaseSession | None, _ifrom: JID) -> str:
        if session is None:
            raise RuntimeError
        roster_iq = await self.xmpp["xep_0356"].get_roster(session.user_jid.bare)

        contacts = session.contacts.known_contacts()

        added = 0
        removed = 0
        updated = 0
        for item in roster_iq["roster"]:
            groups = set(item["groups"])
            if self.xmpp.ROSTER_GROUP in groups:
                contact = contacts.pop(item["jid"], None)
                if contact is None:
                    if len(groups) == 1:
                        await self.xmpp["xep_0356"].set_roster(
                            session.user_jid, {item["jid"]: {"subscription": "remove"}}
                        )
                        removed += 1
                    else:
                        groups.remove(self.xmpp.ROSTER_GROUP)
                        await self.xmpp["xep_0356"].set_roster(
                            session.user_jid,
                            {
                                item["jid"]: {
                                    "subscription": item["subscription"],
                                    "name": item["name"],
                                    "groups": groups,
                                }
                            },
                        )
                        updated += 1
                else:
                    if contact.name != item["name"]:
                        await contact.add_to_roster(force=True)
                        updated += 1

        # we popped before so this only acts on slidge contacts not in the xmpp roster
        for contact in contacts.values():
            added += 1
            await contact.add_to_roster()

        return f"{added} added, {removed} removed, {updated} updated"


class ListContacts(Command):
    NAME = HELP = "👤 List your legacy contacts"
    CHAT_COMMAND = "contacts"
    NODE = CONTACTS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = CONTACTS

    async def run(self, session: AnyBaseSession | None, _ifrom: JID, *_) -> TableResult:
        assert session is not None
        await session.ready
        contacts = sorted(
            session.contacts, key=lambda c: c.name.casefold() if c.name else ""
        )
        return TableResult(
            description="Your buddies",
            fields=[FormField("name"), FormField("jid", type="jid-single")],
            items=[{"name": c.name, "jid": c.jid.bare} for c in contacts],
        )


class ListGroups(Command):
    NAME = HELP = "👥 List your legacy groups"
    CHAT_COMMAND = "groups"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = GROUPS

    async def run(self, session, _ifrom, *_):
        assert session is not None
        await session.ready
        groups = sorted(
            session.bookmarks, key=lambda g: (g.name or g.jid.node).casefold()
        )
        return TableResult(
            description="Your groups",
            fields=[FormField("name"), FormField("jid", type="jid-single")],
            items=[{"name": g.name, "jid": g.jid.bare} for g in groups],
            jids_are_mucs=True,
        )


class Login(Command):
    NAME = "🔐 Re-login to the legacy network"
    HELP = "Login to the legacy service"
    CHAT_COMMAND = "re-login"
    NODE = "https://slidge.im/command/core/" + CHAT_COMMAND

    ACCESS = CommandAccess.USER_NON_LOGGED

    async def run(self, session: AnyBaseSession | None, _ifrom, *_):
        assert session is not None
        if session.is_logging_in:
            raise XMPPError("bad-request", "You are already logging in.")
        session.is_logging_in = True
        try:
            msg = await self.xmpp.login_wrap(session)
        except Exception as e:
            session.send_gateway_status(f"Re-login failed: {e}", show="dnd")
            raise XMPPError(
                "internal-server-error", etype="wait", text=f"Could not login: {e}"
            )
        finally:
            session.is_logging_in = False
        session.logged = True
        session.send_gateway_status(msg or "Re-connected", show="chat")
        session.send_gateway_message(msg or "Re-connected")
        return msg


class CreateGroup(Command):
    NAME = "🆕 New legacy group"
    HELP = "Create a group on the legacy service"
    CHAT_COMMAND = "create-group"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    CATEGORY = GROUPS

    ACCESS = CommandAccess.USER_LOGGED

    async def run(self, session: AnyBaseSession | None, _ifrom, *_):
        assert session is not None
        await session.ready
        contacts = session.contacts.known_contacts(only_friends=True)
        return Form(
            title="Create a new group",
            instructions="Pick contacts that should be part of this new group",
            fields=[
                FormField(var="group_name", label="Name of the group", required=True),
                FormField(
                    var="contacts",
                    label="Contacts to add to the new group",
                    type="list-multi",
                    options=[
                        {"value": str(contact.jid), "label": contact.name}
                        for contact in sorted(contacts.values(), key=lambda c: c.name)
                    ],
                    required=False,
                ),
            ],
            handler=self.finish,
        )

    @staticmethod
    async def finish(form_values: FormValues, session: AnyBaseSession | None, *_):
        assert session is not None
        legacy_id: LegacyGroupIdType = await session.on_create_group(  # type:ignore
            cast(str, form_values["group_name"]),
            [
                await session.contacts.by_jid(JID(j))
                for j in form_values.get("contacts", [])  # type:ignore
            ],
        )
        muc = await session.bookmarks.by_legacy_id(legacy_id)
        return TableResult(
            description=f"Your new group: xmpp:{muc.jid}?join",
            fields=[FormField("name"), FormField("jid", type="jid-single")],
            items=[{"name": muc.name, "jid": muc.jid}],
            jids_are_mucs=True,
        )


class Preferences(Command):
    NAME = "⚙️ Preferences"
    HELP = "Customize the gateway behaviour to your liking"
    CHAT_COMMAND = "preferences"
    NODE = "https://slidge.im/command/core/preferences"
    ACCESS = CommandAccess.USER

    async def run(self, session: AnyBaseSession | None, _ifrom: JID, *_: Any) -> Form:
        fields = deepcopy(self.xmpp.PREFERENCES)
        assert session is not None
        current = session.user.preferences
        for field in fields:
            field.value = current.get(field.var, field.value)  # type:ignore
        return Form(
            title="Preferences",
            instructions=self.HELP,
            fields=fields,
            handler=self.finish,  # type:ignore
            handler_kwargs={"previous": current},
        )

    async def finish(
        self,
        form_values: UserPreferences,
        session: AnyBaseSession | None,
        *_,
        previous,
    ) -> str:
        assert session is not None
        if previous == form_values:
            return "No preference was changed"

        user = session.user
        user.preferences.update(form_values)  # type:ignore
        self.xmpp.store.users.update(user)

        try:
            await session.on_preferences(previous, form_values)  # type:ignore[arg-type]
        except NotImplementedError:
            pass

        if not previous["sync_avatar"] and form_values["sync_avatar"]:
            await self.xmpp.fetch_user_avatar(session)
        else:
            user.avatar_hash = None

        return "Your preferences have been updated."


class Unregister(Command):
    NAME = "❌ Unregister from the gateway"
    HELP = "Unregister from the gateway"
    CHAT_COMMAND = "unregister"
    NODE = "https://slidge.im/command/core/unregister"
    ACCESS = CommandAccess.USER

    async def run(
        self, session: AnyBaseSession | None, _ifrom: JID, *_: Any
    ) -> Confirmation:
        return Confirmation(
            prompt=f"Are you sure you want to unregister from '{self.xmpp.boundjid}'?",
            success=f"You are not registered to '{self.xmpp.boundjid}' anymore.",
            handler=self.unregister,
        )

    async def unregister(self, session: AnyBaseSession | None, _ifrom: JID) -> str:
        assert session is not None
        await self.xmpp.unregister_user(session.user)
        return "You are not registered anymore. Bye!"


class LeaveGroup(Command):
    NAME = HELP = "❌ Leave a legacy group"
    CHAT_COMMAND = "leave-group"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = GROUPS

    async def run(self, session, _ifrom, *_):
        assert session is not None
        await session.ready
        groups = sorted(session.bookmarks, key=lambda g: g.DISCO_NAME.casefold())
        return Form(
            title="Leave a group",
            instructions="Select the group you want to leave",
            fields=[
                FormField(
                    "group",
                    "Group name",
                    type="list-single",
                    options=[
                        {"label": g.name, "value": str(i)} for i, g in enumerate(groups)
                    ],
                )
            ],
            handler=self.confirm,  # type:ignore
            handler_args=(groups,),
        )

    async def confirm(
        self,
        form_values: FormValues,
        _session: AnyBaseSession,
        _ifrom,
        groups: list[LegacyMUC],
    ):
        group = groups[int(form_values["group"])]  # type:ignore
        return Confirmation(
            prompt=f"Are you sure you want to leave the group '{group.name}'?",
            handler=self.finish,  # type:ignore
            handler_args=(group,),
        )

    @staticmethod
    async def finish(session: AnyBaseSession, _ifrom, group: LegacyMUC) -> None:
        await session.on_leave_group(group.legacy_id)
        await session.bookmarks.remove(group, reason="You left this group via slidge.")


class InviteInGroups(Command):
    NAME = "💌 Re-invite me in my groups"
    HELP = "Ask the gateway to send invitations for all your private groups"
    CHAT_COMMAND = "re-invite"
    NODE = GROUPS.node + "/" + CHAT_COMMAND
    ACCESS = CommandAccess.USER_LOGGED
    CATEGORY = GROUPS

    async def run(self, session, _ifrom, *_):
        assert session is not None
        await session.ready
        for muc in session.bookmarks:
            if muc.type == MucType.GROUP:
                session.send_gateway_invite(
                    muc, reason="You asked to be re-invited in all groups."
                )



"""Fix missing existing constraints in SQLite

Revision ID: e365e04e9ac3
Revises: abea9b63fa34
Create Date: 2026-02-11 19:57:06.448033

"""

import logging
from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "e365e04e9ac3"
down_revision: str | None = "abea9b63fa34"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    fix_constraints("contact", "user_account_id", "jid", "legacy_id")
    fix_constraints("contact_sent", "contact_id", "msg_id")
    fix_constraints("mam", "room_id", "stanza_id")
    fix_constraints("attachment", "user_account_id", "legacy_file_id")
    fix_constraints("participant", "room_id", "resource", "contact_id", "occupant_id")


def fix_constraints(table: str, col1: str, *cols: str):
    bogus_name = f"uq_{table}_{col1}"

    inspector = sa.inspect(op.get_bind())
    constraints = inspector.get_unique_constraints(table)
    constraint_exists = any(
        constraint["name"] == bogus_name for constraint in constraints
    )
    if not constraint_exists:
        # If this runs on a newly created DB, this is unnecessary because we actually
        # edited past migration to allow running with postgresql.
        logging.info(f"No need to fix constraints for {bogus_name}")
        return

    with op.batch_alter_table(table, schema=None) as batch_op:
        logging.info(f"Dropping constraint {bogus_name}")
        batch_op.drop_constraint(bogus_name, type_="unique")
        for col2 in cols:
            op.execute(f"""
                DELETE FROM '{table}'
                WHERE id NOT IN (
                    SELECT MIN(id)
                    FROM '{table}'
                    GROUP BY '{col1}', '{col2}'
                );
            """)
            proper_name = f"uq_{table}_{col1}_{col2}"
            logging.info(f"Creating constraint {proper_name}")
            batch_op.create_unique_constraint(proper_name, [col1, col2])


def downgrade() -> None:
    raise RuntimeError("Downgrade not supported!")



from copy import copy

from slixmpp import CoroutineCallback, Iq, StanzaPath, register_stanza_plugin
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0084 import MetaData
from slixmpp.plugins.xep_0292.stanza import NS as VCard4NS

from ...contact import LegacyContact
from ...core.session import BaseSession
from ...group import LegacyParticipant
from ...util.types import Avatar
from .util import DispatcherMixin, exceptions_to_xmpp_errors


class VCardMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp) -> None:
        super().__init__(xmpp)
        xmpp.register_handler(
            CoroutineCallback(
                "get_vcard", StanzaPath("iq@type=get/vcard"), self.on_get_vcard
            )
        )
        xmpp.remove_handler("VCardTemp")
        xmpp.register_handler(
            CoroutineCallback(
                "VCardTemp",
                StanzaPath("iq/vcard_temp"),
                self.__vcard_temp_handler,
            )
        )
        # TODO: MR to slixmpp adding this to XEP-0084
        register_stanza_plugin(
            self.xmpp.plugin["xep_0060"].stanza.Item,
            MetaData,
        )

    @exceptions_to_xmpp_errors
    async def on_get_vcard(self, iq: Iq) -> None:
        session = await self._get_session(iq, logged=True)
        contact = await session.contacts.by_jid(iq.get_to())
        vcard = await contact.get_vcard()
        reply = iq.reply()
        if vcard:
            reply.append(vcard)
        else:
            reply.enable("vcard")
        reply.send()

    @exceptions_to_xmpp_errors
    async def __vcard_temp_handler(self, iq: Iq):
        if iq["type"] == "get":
            return await self.__handle_get_vcard_temp(iq)

        if iq["type"] == "set":
            return await self.__handle_set_vcard_temp(iq)

    async def __fetch_user_avatar(self, session: BaseSession):
        hash_ = session.user.avatar_hash
        if not hash_:
            raise XMPPError(
                "item-not-found", "The slidge user does not have any avatar set"
            )
        meta_iq = await self.xmpp.plugin["xep_0060"].get_item(
            session.user_jid,
            MetaData.namespace,
            hash_,
            ifrom=self.xmpp.boundjid.bare,
        )
        info = meta_iq["pubsub"]["items"]["item"]["avatar_metadata"]["info"]
        type_ = info["type"]
        data_iq = await self.xmpp.plugin["xep_0084"].retrieve_avatar(
            session.user_jid, hash_, ifrom=self.xmpp.boundjid.bare
        )
        bytes_ = data_iq["pubsub"]["items"]["item"]["avatar_data"]["value"]
        return bytes_, type_

    async def __handle_get_vcard_temp(self, iq: Iq):
        session = self.xmpp.get_session_from_stanza(iq)
        entity = await session.get_contact_or_group_or_participant(iq.get_to(), False)
        if not entity:
            raise XMPPError("item-not-found")

        bytes_ = None
        if isinstance(entity, LegacyParticipant):
            if entity.is_user:
                bytes_, type_ = await self.__fetch_user_avatar(session)
                if not bytes_:
                    raise XMPPError(
                        "internal-server-error",
                        "Could not fetch the slidge user's avatar",
                    )
                avatar = None
                vcard = None
            elif not (contact := entity.contact):
                raise XMPPError("item-not-found", "This participant has no contact")
            else:
                vcard = await contact.get_vcard()
                avatar = contact.get_avatar()
                type_ = "image/png"
        else:
            avatar = entity.get_avatar()
            type_ = "image/png"
            if isinstance(entity, LegacyContact):
                vcard = await entity.get_vcard()
            else:
                vcard = None
        v = self.xmpp.plugin["xep_0054"].make_vcard()
        if avatar is not None and avatar.data:
            bytes_ = avatar.data.get_value()
        if bytes_:
            v["PHOTO"]["BINVAL"] = bytes_
            v["PHOTO"]["TYPE"] = type_
        if vcard:
            for el in vcard.xml:
                new = copy(el)
                new.tag = el.tag.replace(f"{{{VCard4NS}}}", "")
                v.append(new)
        reply = iq.reply()
        reply.append(v)
        reply.send()

    async def __handle_set_vcard_temp(self, iq: Iq):
        muc = await self.get_muc_from_stanza(iq)
        to = iq.get_to()

        if to.resource:
            raise XMPPError("bad-request", "You cannot set participants avatars")

        data = iq["vcard_temp"]["PHOTO"]["BINVAL"] or None
        try:
            legacy_id = await muc.on_avatar(
                data, iq["vcard_temp"]["PHOTO"]["TYPE"] or None
            )
        except XMPPError:
            raise
        except Exception as e:
            raise XMPPError("internal-server-error", str(e))
        reply = iq.reply(clear=True)
        reply.enable("vcard_temp")
        reply.send()

        if not data:
            await muc.set_avatar(None)
            return

        if legacy_id:
            await muc.set_avatar(Avatar(data=data, unique_id=legacy_id))



"""
Handling groups
"""

import uuid
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from slidge import LegacyBookmarks, LegacyMUC, LegacyParticipant, MucType
from slidge.util.types import Avatar, Hat, HoleBound

if TYPE_CHECKING:
    from .session import Session


class Bookmarks(LegacyBookmarks):
    async def fill(self):
        for i in "aaa", "bbb":
            muc = await self.by_legacy_id(i)
            await muc.add_to_bookmarks()


class MUC(LegacyMUC):
    session: "Session"
    type = MucType.GROUP

    async def update_info(self):
        info = await self.session.legacy_client.get_group_info(self.legacy_id)
        self.name = info.name
        await self.set_avatar(Avatar(path=info.avatar, unique_id=info.avatar_unique_id))

    async def fill_participants(self):
        # in a real case, this would probably call something like
        # self.session.legacy_client.fetch_group_members(self.legacy_id)
        for i in 0, 111, 222:
            part = await self.get_participant_by_legacy_id(i)
            if i == 111:
                part.role = "moderator"
                part.affiliation = "owner"
                part.set_hats([Hat("test", "test"), Hat("prout", "prout")])
            yield part
        me = await self.get_user_participant()
        me.role = "moderator"
        me.affiliation = "owner"
        yield me

    async def backfill(
        self,
        after: HoleBound | None = None,
        before: HoleBound | None = None,
    ):
        # in a real case, this would probably call something like
        # self.session.legacy_client.fetch_group_history(self.legacy_id)
        for i in range(10):
            part = await self.get_participant_by_legacy_id(0)
            part.send_text(
                f"History message #{i}",
                when=datetime.now() - timedelta(hours=i),
                legacy_msg_id=f"{i}--{uuid.uuid4().hex}",
                archive_only=True,
            )


class Participant(LegacyParticipant):
    pass



"""
This module implements an unified API to define :term:`adhoc <Ad-hoc Command>`
or :term:`chatbot <Chatbot Command>` commands. Just subclass a :class:`Command`,
and make sures it is imported in your legacy module's ``__init__.py``.
"""

from . import admin, register, user  # noqa: F401
from .base import (
    Command,
    CommandAccess,
    CommandResponseType,
    Confirmation,
    Form,
    FormField,
    SearchResult,
    TableResult,
)

__all__ = (
    "Command",
    "CommandAccess",
    "CommandResponseType",
    "Confirmation",
    "Form",
    "FormField",
    "SearchResult",
    "TableResult",
)



import pytest
import slixmpp.test
from slixmpp import JID

import slidge.command.chat_command
from slidge.command import Command, Confirmation
from slidge.slixfix.delivery_receipt import DeliveryReceipt
from slidge.util.test import SlixTestPlus


class MockSession:
    def __init__(self, jid):
        self.logged = "logged" in jid.user


@pytest.fixture(autouse=True)
def mock(monkeypatch, MockRE):
    monkeypatch.setattr(Command, "_get_session", lambda s, j: MockSession(j))
    monkeypatch.setattr(
        slixmpp.test.ComponentXMPP,
        "get_session_from_stanza",
        lambda self, stanza: MockSession(stanza.get_from()),
        raising=False,
    )
    monkeypatch.setattr(
        slixmpp.test.ComponentXMPP,
        "jid_validator",
        MockRE,
        raising=False,
    )
    monkeypatch.setattr(
        slixmpp.test.ComponentXMPP,
        "get_session_from_jid",
        lambda s, j: MockSession(j),
        raising=False,
    )

class CommandAdmin(Command):
    NAME = "Command number one"
    CHAT_COMMAND = "command1"

    test_results = []

    async def run(self, _session, _ifrom):
        return Confirmation(
            prompt="Confirm?", handler=self.finish, success="It worked!"
        )

    async def finish(self, _session, _ifrom):
        self.test_results.append("yup")


class CommandAdminConfirmFail(CommandAdmin):
    NAME = "Command number two"
    CHAT_COMMAND = "command2"

    async def run_admin(self):
        return Confirmation(
            prompt="Confirm?", handler=self.finish, success="It worked!"
        )

    async def finish(self, _session, _ifrom):
        raise RuntimeError("Ploup")


class TestChatCommands(SlixTestPlus):
    def setUp(self):
        self.stream_start(
            mode="component",
            plugins=["xep_0050"],
            jid="slidge.whatever.ass",
            server="whatever.ass",
        )
        self.commands = slidge.command.chat_command.ChatCommandProvider(self.xmpp)
        self.commands.register(CommandAdmin(self.xmpp))
        self.commands.register(CommandAdminConfirmFail(self.xmpp))
        self.xmpp.delivery_receipt = DeliveryReceipt(self.xmpp)
        super().setUp()

    def test_non_existing(self):
        self.recv(  # language=XML
            f"""
            <message from='admin@whatever.ass/cheogram'
                     to='{self.xmpp.boundjid.bare}'
                     type='chat'
                     id='not-found'>
              <body>non-existing</body>
            </message>
            """
        )
        t = self.commands.UNKNOWN.format("non-existing")
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     from="slidge.whatever.ass"
                     to="admin@whatever.ass/cheogram"
                     type="chat">
              <body>{t}</body>
            </message>
            """
        )
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     from="slidge.whatever.ass"
                     to="admin@whatever.ass/cheogram"
                     type="error"
                     id='not-found'>
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">{t}</text>
              </error>
            </message>
            """,
            use_values=False,
        )

    def test_other_destination(self):
        self.recv(  # language=XML
            f"""
            <message from='admin@whatever.ass/cheogram'
                     to='something@{self.xmpp.boundjid.bare}'
                     type='chat'
                     id='not-found'>
              <body>help</body>
            </message>
            """
        )
        assert self.next_sent() is None

    def test_command_help(self):
        self.recv(  # language=XML
            f"""
            <message from='admin@whatever.ass/cheogram'
                     to='{self.xmpp.boundjid.bare}'
                     type='chat'
                     id='help'>
              <body>help</body>
            </message>
            """
        )
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     from="slidge.whatever.ass"
                     to="admin@whatever.ass/cheogram"
                     type="chat">
              <body>Available commands:\ncommand1 -- Command number one\ncommand2 -- Command number two</body>
            </message>
            """
        )

    def test_input(self):
        fut = self.run_coro(
            self.commands.input(JID("user@whatever.ass/x"), "blabla", blocking=False)
        )

        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     type="chat"
                     to="user@whatever.ass/x"
                     from="slidge.whatever.ass">
              <body>blabla</body>
            </message>
            """
        )
        self.recv(  # language=XML
            """
            <message from='user@whatever.ass/y'
                     to='slidge.whatever.ass'>
              <body>reply</body>
            </message>
            """
        )

        assert fut.result() == "reply"

    def test_confirm_no(self):
        self.recv(  # language=XML
            f"""
            <message from='admin@whatever.ass/cheogram'
                     to='{self.xmpp.boundjid.bare}'
                     type='chat'
                     id='help'>
              <body>command1</body>
            </message>
            """
        )
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     type="chat"
                     to="admin@whatever.ass/cheogram"
                     from="{self.xmpp.boundjid.bare}">
              <body>Confirm?</body>
            </message>
            """
        )
        self.recv(  # language=XML
            f"""
            <message from='admin@whatever.ass/cheogram'
                     to='{self.xmpp.boundjid.bare}'
                     type='chat'
                     id='help'>
              <body>no</body>
            </message>
            """
        )
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     type="chat"
                     to="admin@whatever.ass/cheogram"
                     from="{self.xmpp.boundjid.bare}">
              <body>Canceled</body>
            </message>
            """
        )
        assert len(CommandAdmin.test_results) == 0

    def test_confirm_yes(self):
        self.recv(  # language=XML
            f"""
            <message from='admin@whatever.ass/cheogram'
                     to='{self.xmpp.boundjid.bare}'
                     type='chat'
                     id='help'>
              <body>command1</body>
            </message>
            """
        )
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     type="chat"
                     to="admin@whatever.ass/cheogram"
                     from="{self.xmpp.boundjid.bare}">
              <body>Confirm?</body>
            </message>
            """
        )
        self.recv(  # language=XML
            f"""
            <message from='admin@whatever.ass/cheogram'
                     to='{self.xmpp.boundjid.bare}'
                     type='chat'
                     id='help'>
              <body>yes</body>
            </message>
            """
        )
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     type="chat"
                     to="admin@whatever.ass/cheogram"
                     from="{self.xmpp.boundjid.bare}">
              <body>End of command.</body>
            </message>
            """
        )
        assert CommandAdmin.test_results.pop() == "yup"
        assert len(CommandAdmin.test_results) == 0



import re
from datetime import datetime, timezone
from pathlib import Path

import pytest

from slixmpp import JID

from slidge.contact import LegacyContact
from slidge.core import config
from slidge.core.mixins.presence import get_last_seen_fallback
from slidge.util import (
    SubclassableOnce,
    is_valid_phone_number,
)
from slidge.util.types import Mention, LegacyAttachment
from slidge.util.util import merge_resources, replace_mentions, strip_leading_emoji


def test_subclass():
    SubclassableOnce.TEST_MODE = False
    # fmt: off
    class A(SubclassableOnce): pass

    with pytest.raises(AttributeError):
        A.get_unique_subclass()

    assert A.get_self_or_unique_subclass() is A

    class B(A): pass
    assert A.get_self_or_unique_subclass() is B

    with pytest.raises(AttributeError):
        B.get_unique_subclass()

    with pytest.raises(RuntimeError):
        class C(A): pass

    A.reset_subclass()

    class C(A): pass
    assert A.get_self_or_unique_subclass() is C

    A.reset_subclass()
    class D(SubclassableOnce): pass

    assert D.get_self_or_unique_subclass() is D

    # fmt: on
    SubclassableOnce.TEST_MODE = True


def test_phone_validation():
    assert is_valid_phone_number("+33")
    assert not is_valid_phone_number("+")
    assert not is_valid_phone_number("+asdfsadfa48919sadf")
    assert not is_valid_phone_number("12597891")


def test_strip_delay(monkeypatch):
    monkeypatch.setattr(config, "IGNORE_DELAY_THRESHOLD", 300)

    class MockDelay:
        @staticmethod
        def set_stamp(x):
            pass

        @staticmethod
        def set_from(x):
            pass

    class MockC:
        STRIP_SHORT_DELAY = True

        class xmpp:
            boundjid = JID("test")

    class MockMsg:
        delay_added = None

        def __getitem__(self, key):
            if key == "delay":
                self.delay_added = True
            return MockDelay

    msg = MockMsg()
    LegacyContact._add_delay(MockC, msg, datetime.now())
    assert not msg.delay_added

    monkeypatch.setattr(config, "IGNORE_DELAY_THRESHOLD", 0)

    msg = MockMsg()
    LegacyContact._add_delay(MockC, msg, datetime.now())
    assert msg.delay_added


def test_merge_presence():
    assert merge_resources(
        {
            "1": {
                "show": "",
                "status": "",
                "priority": 0,
            }
        }
    ) == {
        "show": "",
        "status": "",
        "priority": 0,
    }

    assert merge_resources(
        {
            "1": {
                "show": "dnd",
                "status": "X",
                "priority": -10,
            },
            "2": {
                "show": "dnd",
                "status": "",
                "priority": 0,
            },
        }
    ) == {
        "show": "dnd",
        "status": "X",
        "priority": 0,
    }

    assert merge_resources(
        {
            "1": {
                "show": "",
                "status": "",
                "priority": 0,
            },
            "2": {
                "show": "away",
                "status": "",
                "priority": 0,
            },
            "3": {
                "show": "dnd",
                "status": "",
                "priority": 0,
            },
        }
    ) == {
        "show": "",
        "status": "",
        "priority": 0,
    }

    assert merge_resources(
        {
            "1": {
                "show": "",
                "status": "",
                "priority": 0,
            },
            "2": {
                "show": "away",
                "status": "",
                "priority": 0,
            },
            "3": {
                "show": "dnd",
                "status": "Blah blah",
                "priority": 0,
            },
        }
    ) == {
        "show": "",
        "status": "Blah blah",
        "priority": 0,
    }

    assert merge_resources(
        {
            "1": {
                "show": "",
                "status": "",
                "priority": 0,
            },
            "2": {
                "show": "away",
                "status": "Blah",
                "priority": 0,
            },
            "3": {
                "show": "dnd",
                "status": "Blah blah",
                "priority": 10,
            },
        }
    ) == {
        "show": "",
        "status": "Blah blah",
        "priority": 0,
    }

    assert merge_resources(
        {
            "1": {
                "show": "",
                "status": "",
                "priority": 0,
            },
            "2": {
                "show": "away",
                "status": "Blah",
                "priority": 0,
            },
            "3": {
                "show": "dnd",
                "status": "",
                "priority": 10,
            },
        }
    ) == {
        "show": "",
        "status": "Blah",
        "priority": 0,
    }


def test_replace_mentions():
    mentions = []
    text = "Text Mention 1 and Mention 2, and Mention 3"
    for match in re.finditer("Mention 1|Mention 2|Mention 3", text):
        span = match.span()
        nick = match.group()
        mentions.append(
            Mention(contact=f"Contact{nick[-1]}", start=span[0], end=span[1])
        )
    assert (
        replace_mentions(text, mentions, lambda c: "@" + c[-1])
        == "Text @1 and @2, and @3"
    )

    mentions = []
    text = "Text Mention 1 and Mention 2, and Mention 3 blabla"
    for match in re.finditer("Mention 1|Mention 2|Mention 3", text):
        span = match.span()
        nick = match.group()
        mentions.append(
            Mention(contact=f"Contact{nick[-1]}", start=span[0], end=span[1])
        )
    assert (
        replace_mentions(text, mentions, lambda c: "@" + c[-1])
        == "Text @1 and @2, and @3 blabla"
    )


def test_strip_emoji():
    try:
        import emoji
    except ImportError:
        return
    assert strip_leading_emoji("🛷️ Slidge administration") == "Slidge administration"
    assert strip_leading_emoji("no emoji") == "no emoji"
    assert strip_leading_emoji("no") == "no"
    assert strip_leading_emoji("👤 Contacts") == "Contacts"
    assert strip_leading_emoji("👥 Groups") == "Groups"


def test_last_seen_fallback_formatting():
    now = datetime.now(tz=timezone.utc)
    res, should_update = get_last_seen_fallback(now)
    assert should_update
    assert "AM" in res or "PM" in res


def test_attachment_str():
    att = LegacyAttachment(path="whatevs")
    assert str(att) == f"Attachment(path={Path('whatevs')!r})"

    att = LegacyAttachment(url="https://url", legacy_file_id="xxx")
    assert str(att) == "Attachment(legacy_file_id='xxx', url='https://url')"

    att = LegacyAttachment(data=b"aaaaa", legacy_file_id="xxx")
    assert str(att) == "Attachment(legacy_file_id='xxx', data=<5 bytes>)"

    att = LegacyAttachment(data=b"aaaaa")
    assert str(att) == "Attachment(data=<5 bytes>)"



import logging
from typing import TYPE_CHECKING

from slixmpp import Message
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.plugins.xep_0084.stanza import Info

from ..session import BaseSession
from .caps import CapsMixin
from .disco import DiscoMixin
from .message import MessageMixin
from .muc import MucMixin
from .presence import PresenceHandlerMixin
from .registration import RegistrationMixin
from .search import SearchMixin
from .util import exceptions_to_xmpp_errors
from .vcard import VCardMixin

if TYPE_CHECKING:
    from slidge.core.gateway import BaseGateway


class SessionDispatcher(
    CapsMixin,
    DiscoMixin,
    RegistrationMixin,
    MessageMixin,
    MucMixin,
    PresenceHandlerMixin,
    SearchMixin,
    VCardMixin,
):
    __slots__: list[str] = ["xmpp", "_MamMixin__mam_cleanup_task"]

    def __init__(self, xmpp: "BaseGateway") -> None:
        super().__init__(xmpp)
        xmpp.add_event_handler(
            "avatar_metadata_publish", self.on_avatar_metadata_publish
        )

    @exceptions_to_xmpp_errors
    async def on_avatar_metadata_publish(self, m: Message) -> None:
        session = await self._get_session(m, timeout=None)
        if not session.user.preferences.get("sync_avatar", False):
            session.log.debug("User does not want to sync their avatar")
            return
        info = m["pubsub_event"]["items"]["item"]["avatar_metadata"]["info"]

        await self.on_avatar_metadata_info(session, info)

    async def on_avatar_metadata_info(self, session: BaseSession, info: Info) -> None:
        hash_ = info["id"]

        if session.user.avatar_hash == hash_:
            session.log.debug("We already know this avatar hash")
            return

        if hash_:
            try:
                iq = await self.xmpp.plugin["xep_0084"].retrieve_avatar(
                    session.user_jid, hash_, ifrom=self.xmpp.boundjid.bare
                )
            except (IqError, IqTimeout) as e:
                session.log.warning("Could not fetch the user's avatar: %s", e)
                return
            bytes_ = iq["pubsub"]["items"]["item"]["avatar_data"]["value"]
            type_ = info["type"]
            height = info["height"]
            width = info["width"]
        else:
            with self.xmpp.store.session(expire_on_commit=False) as orm:
                session.user.avatar_hash = None
                orm.add(session.user)
                orm.commit()
            bytes_ = type_ = height = width = hash_ = None
        try:
            await session.on_avatar(bytes_, hash_, type_, width, height)
        except NotImplementedError:
            pass
        except Exception as e:
            # If something goes wrong here, replying an error stanza will to the
            # avatar update will likely not show in most clients, so let's send
            # a normal message from the component to the user.
            session.send_gateway_message(
                f"Something went wrong trying to set your avatar: {e!r}"
            )
        else:
            session.user.avatar_hash = hash_
            with self.xmpp.store.session(expire_on_commit=False) as orm:
                orm.add(session.user)
                orm.commit()
            for room in session.bookmarks:
                participant = await room.get_user_participant()
                participant.send_last_presence(force=True, no_cache_online=True)


log = logging.getLogger(__name__)



import logging
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import TYPE_CHECKING, Any, TypeVar

from slixmpp import JID, Iq, Message, Presence
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import StanzaBase

from ...contact.roster import ContactIsUser
from ...util.types import LegacyMessageType, Recipient, RecipientType
from ..session import BaseSession

if TYPE_CHECKING:
    from slidge import BaseGateway
    from slidge.group import LegacyMUC


class Ignore(BaseException):
    pass


class DispatcherMixin:
    __slots__: list[str] = []

    def __init__(self, xmpp: "BaseGateway") -> None:
        self.xmpp = xmpp  # type:ignore[misc]

    async def _get_session(
        self,
        stanza: Message | Presence | Iq,
        timeout: int | None = 10,
        wait_for_ready: bool = True,
        logged: bool = False,
    ) -> BaseSession:
        xmpp = self.xmpp
        if stanza.get_from().server == xmpp.boundjid.bare:
            log.debug("Ignoring echo")
            raise Ignore
        if (
            isinstance(stanza, Message)
            and stanza.get_type() == "chat"
            and stanza.get_to() == xmpp.boundjid.bare
        ):
            log.debug("Ignoring message to component")
            raise Ignore
        session = await self._get_session_from_jid(
            stanza.get_from(), timeout, wait_for_ready, logged
        )
        if isinstance(stanza, Message) and _ignore(session, stanza):
            raise Ignore
        return session

    async def _get_session_from_jid(
        self,
        jid: JID,
        timeout: int | None = 10,
        wait_for_ready: bool = True,
        logged: bool = False,
    ) -> BaseSession:
        session = self.xmpp.get_session_from_jid(jid)
        if session is None:
            raise XMPPError("registration-required")
        if logged:
            session.raise_if_not_logged()
        if wait_for_ready:
            await session.wait_for_ready(timeout)
        return session

    async def get_muc_from_stanza(self, iq: Iq | Message | Presence) -> "LegacyMUC":
        ito = iq.get_to()
        if ito == self.xmpp.boundjid.bare:
            raise XMPPError("bad-request", text="This is only handled for MUCs")

        session = await self._get_session(iq, logged=True)
        muc = await session.bookmarks.by_jid(ito)
        return muc

    def _xmpp_msg_id_to_legacy(
        self,
        session: "BaseSession[LegacyMessageType, Any]",
        xmpp_id: str,
        recipient: Recipient,
        origin: bool = False,
    ) -> LegacyMessageType:
        with self.xmpp.store.session() as orm:
            sent = self.xmpp.store.id_map.get_legacy(
                orm, recipient.stored.id, xmpp_id, recipient.is_group, origin
            )
            if sent is not None:
                return self.xmpp.LEGACY_MSG_ID_TYPE(sent)

        try:
            return session.xmpp_to_legacy_msg_id(xmpp_id)
        except XMPPError:
            raise
        except Exception as e:
            log.debug("Couldn't convert xmpp msg ID to legacy ID.", exc_info=e)
            raise XMPPError(
                "internal-server-error", "Couldn't convert xmpp msg ID to legacy ID."
            )

    async def _get_session_recipient_thread(
        self, msg: Message
    ) -> tuple["BaseSession", Recipient, int | str]:
        session = await self._get_session(msg)
        e: Recipient = await get_recipient(session, msg)
        legacy_thread = await self._xmpp_to_legacy_thread(session, msg, e)
        return session, e, legacy_thread

    async def _xmpp_to_legacy_thread(
        self, session: "BaseSession", msg: Message, recipient: RecipientType
    ):
        xmpp_thread = msg["thread"]
        if not xmpp_thread:
            return None

        if session.MESSAGE_IDS_ARE_THREAD_IDS:
            return self._xmpp_msg_id_to_legacy(session, xmpp_thread, recipient)

        with session.xmpp.store.session() as orm:
            legacy_thread_str = session.xmpp.store.id_map.get_thread(
                orm, recipient.stored.id, xmpp_thread, recipient.is_group
            )
            if legacy_thread_str is not None:
                return session.xmpp.LEGACY_MSG_ID_TYPE(legacy_thread_str)
        async with session.thread_creation_lock:
            legacy_thread = await recipient.create_thread(xmpp_thread)
            with session.xmpp.store.session() as orm:
                session.xmpp.store.id_map.set_thread(
                    orm,
                    recipient.stored.id,
                    str(legacy_thread),
                    xmpp_thread,
                    recipient.is_group,
                )
                orm.commit()
            return legacy_thread


def _ignore(session: "BaseSession", msg: Message) -> bool:
    i = msg.get_id()
    if i.startswith("slidge-carbon-"):
        return True
    if i not in session.ignore_messages:
        return False
    session.log.debug("Ignored sent carbon: %s", i)
    session.ignore_messages.remove(i)
    return True


async def get_recipient(session: "BaseSession", m: Message) -> RecipientType:
    session.raise_if_not_logged()
    if m.get_type() == "groupchat":
        muc = await session.bookmarks.by_jid(m.get_to())
        r = m.get_from().resource
        if r not in muc.get_user_resources():
            session.create_task(muc.kick_resource(r))
            raise XMPPError("not-acceptable", "You are not connected to this chat")
        return muc
    else:
        return await session.contacts.by_jid(m.get_to())


StanzaType = TypeVar("StanzaType", bound=StanzaBase)
HandlerType = Callable[[Any, StanzaType], Awaitable[None]]


def exceptions_to_xmpp_errors(cb: HandlerType) -> HandlerType:
    @wraps(cb)
    async def wrapped(*args):
        try:
            await cb(*args)
        except Ignore:
            pass
        except XMPPError:
            raise
        except NotImplementedError:
            log.debug("NotImplementedError raised in %s", cb)
            raise XMPPError(
                "feature-not-implemented",
                f"{cb.__name__} is not implemented by the legacy module",
                clear=False,
            )
        except ContactIsUser:
            raise XMPPError(
                "bad-request", "Actions with your bridged self are not allowed."
            )
        except Exception as e:
            log.error("Failed to handle incoming stanza: %s", args, exc_info=e)
            raise XMPPError("internal-server-error", str(e))

    return wrapped


log = logging.getLogger(__name__)



import pytest
import sqlalchemy as sa
from slixmpp import JID
from sqlalchemy import select

import slidge.db.store
from slidge import GatewayUser
from slidge.db.meta import Base
from slidge.db.models import Avatar, Contact, Participant, Room
from slidge.db.store import SlidgeStore


@pytest.fixture
def slidge_store(tmp_path):
    engine = sa.create_engine("sqlite+pysqlite:///:memory:", echo=True)
    Base.metadata.create_all(engine)
    import slidge.core.config

    if not hasattr(slidge.core.config, "HOME_DIR"):
        slidge.core.config.HOME_DIR = tmp_path
    yield SlidgeStore(engine)


def test_delete_avatar(slidge_store):
    with slidge_store.session() as orm:
        user = GatewayUser(jid=JID("x@x.com"), legacy_module_data={})
        orm.add(user)
        orm.commit()
        avatar = Avatar(
            hash="hash",
            height=0,
            width=0,
        )

        contact = Contact(
            jid=JID("xxx@xxx.com"), legacy_id="prout", user_account_id=user.id
        )
        orm.add(contact)
        orm.commit()
        contact_pk = contact.id
        contact = slidge_store.contacts.get_by_pk(orm, contact_pk)
        contact.avatar = avatar
        orm.add(contact)

        orm.commit()

    with slidge_store.session() as orm:
        contact = slidge_store.contacts.get_by_pk(orm, contact_pk)
        assert contact.avatar is not None
        orm.delete(contact.avatar)
        orm.commit()
    with slidge_store.session() as orm:
        contact = slidge_store.contacts.get_by_pk(orm, contact_pk)
        assert contact.avatar is None


def test_delete_contact_avatar_orphan(slidge_store):
    with slidge_store.session() as orm:
        user = GatewayUser(jid=JID("x@x.com"), legacy_module_data={})
        orm.add(user)
        orm.commit()

        avatar = Avatar(
            hash="hash",
            height=0,
            width=0,
        )

        contact = Contact(
            jid=JID("xxx@xxx.com"), legacy_id="prout", user_account_id=user.id, avatar=avatar
        )
        orm.add(contact)
        orm.commit()
        avatar_pk = avatar.id

        avatar = orm.get(Avatar, avatar_pk)
        assert avatar is not None

        orm.delete(contact)
        orm.commit()

        avatar = orm.get(Avatar, avatar_pk)
        assert avatar is None


def test_unregister(slidge_store):
    with slidge_store.session() as orm:
        user = GatewayUser(jid=JID("x@x.com"), legacy_module_data={})
        orm.add(user)
        contact = Contact(jid=JID("xxx@xxx.com"), legacy_id="prout", user=user)
        orm.add(contact)
        orm.commit()
        contact_pk = contact.id
        slidge_store.contacts.add_to_sent(orm, contact_pk, "an-id")
        orm.commit()
        orm.delete(user)
        orm.commit()


def test_unregister_with_participants(slidge_store):
    with slidge_store.session() as orm:
        user = GatewayUser(jid=JID("x@x.com"), legacy_module_data={})
        orm.add(user)
        orm.commit()
        contact = Contact(jid=JID("xxx@xxx.com"), legacy_id="prout", user=user)
        orm.add(contact)
        orm.commit()

        room = Room(
            user_account_id=user.id,
            legacy_id="legacy-room",
            jid=JID("legacy-room@something"),
        )
        orm.add(room)
        orm.commit()

        contacts = orm.execute(select(Contact).where(Contact.user_account_id == user.id)).scalars()
        assert len(list(contacts)) == 1
        rooms = orm.execute(select(Room).where(Room.user_account_id == user.id)).scalars()
        assert len(list(rooms)) == 1

        participant = Participant(
            room_id=room.id,
            contact_id=contact.id,
            resource="whatever",
            nickname="whatever",
            nickname_no_illegal="whatever",
            occupant_id="uuid",
        )
        orm.add(participant)
        orm.commit()

        orm.delete(user)
        orm.commit()

        rooms = orm.execute(select(Room).where(Room.user_account_id == user.id)).scalars()
        assert len(list(rooms)) == 0

        contacts = orm.execute(select(Contact).where(Contact.user_account_id == user.id)).scalars()
        assert len(list(contacts)) == 0



import unittest

import pytest
from slixmpp.exceptions import XMPPError

from conftest import AvatarFixtureMixin
from slixmpp import JID, Iq

from slidge import BaseGateway, BaseSession, GatewayUser, LegacyRoster, LegacyBookmarks, LegacyMUC, LegacyContact
from slidge.core.session import _sessions
from slidge.util.test import SlidgeTest
from slidge.util.types import LegacyUserIdType, LegacyMUCType, MucType


class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"
    GROUPS = True


class Contact(LegacyContact):
    async def update_info(self):
        if self.legacy_id.startswith("group"):
            raise XMPPError()


class Session(BaseSession):
    async def login(self):
        return "YUP"


class Bookmarks(LegacyBookmarks):
    async def fill(self) -> None:
        pass


class MUC(LegacyMUC):
    async def update_info(self):
        if not self.legacy_id.startswith("group"):
            raise XMPPError()
        self.type = MucType.GROUP

@pytest.mark.usefixtures("avatar")
class TestMUCSubject(AvatarFixtureMixin, SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@montague.lit/gajim").bare,
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()
        self.run_coro(
            self.xmpp._BaseGateway__dispatcher._on_user_register(
                Iq(sfrom="romeo@montague.lit/gajim")
            )
        )
        welcome = self.next_sent()
        assert welcome["body"]
        stanza = self.next_sent()
        assert "logging in" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing contacts" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing groups" in stanza["status"].lower(), stanza
        probe = self.next_sent()
        assert probe.get_type() == "probe"
        stanza = self.next_sent()
        assert "yup" in stanza["status"].lower(), stanza

        self.send(  # language=XML
            """
            <iq type="get"
                to="romeo@montague.lit"
                id="1"
                from="aim.shakespeare.lit">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <items node="urn:xmpp:avatar:metadata" />
              </pubsub>
            </iq>
            """
        )

    @property
    def romeo_session(self) -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )

    def test_empty_subject(self):
        muc = self.run_coro(self.romeo_session.bookmarks.by_legacy_id("group"))
        with unittest.mock.patch("uuid.uuid4", return_value="uuid"):
            self.recv(  # language=XML
                f"""
            <presence from="romeo@montague.lit/movim"
                      to="{muc.jid}/nick">
              <x xmlns='http://jabber.org/protocol/muc' />
            </presence>
            """            )
            self.send(  # language=XML
                """
            <presence from="group@aim.shakespeare.lit/romeo"
                      to="romeo@montague.lit/movim">
              <x xmlns="http://jabber.org/protocol/muc#user">
                <item affiliation="member"
                      role="participant"
                      jid="romeo@montague.lit/movim" />
                <status code="210" />
                <status code="110" />
                <status code="100" />
              </x>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="slidge-user" />
            </presence>
            """,
            use_values=False)
            self.send(  # language=XML
                """
            <message type="groupchat"
                     from="group@aim.shakespeare.lit"
                     to="romeo@montague.lit/movim">
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="uuid"
                         by="group@aim.shakespeare.lit" />
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="room" />
              <subject />
            </message>
            """,
            use_values=False,
            )

    def test_set_thread_subject(self):
        muc: MUC = self.run_coro(self.romeo_session.bookmarks.by_legacy_id("group"))
        muc.add_user_resource("movim")
        with unittest.mock.patch("uuid.uuid4", return_value="uuid"):
            juliet_participant = self.run_coro(muc.get_participant("juliet"))
            juliet_participant.set_thread_subject("legacy-thread-id", "some-subject")
            self.send(  # language=XML
                """
            <message xmlns="jabber:component:accept"
                     type="groupchat"
                     from="group@aim.shakespeare.lit/juliet"
                     to="romeo@montague.lit/movim">
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="uuid"
                         by="group@aim.shakespeare.lit" />
              <thread>legacy-thread-id</thread>
              <subject>some-subject</subject>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="uuid" />
            </message>
            """
            )

    def test_user_set_thread_subject(self):
        muc: MUC = self.run_coro(self.romeo_session.bookmarks.by_legacy_id("group"))
        muc.add_user_resource("movim")
        with unittest.mock.patch("slidge.group.room.LegacyMUC.on_set_subject") as on_set_subject, unittest.mock.patch("slidge.group.room.LegacyMUC.on_set_thread_subject") as on_set_thread_subject:
            self.recv(  # language=XML
                """
            <message xmlns="jabber:component:accept"
                     type="groupchat"
                     from="romeo@montague.lit/movim"
                     to="group@aim.shakespeare.lit">
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="uuid"
                         by="group@aim.shakespeare.lit" />
              <thread>thread-id</thread>
              <subject>some-subject</subject>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="uuid" />
            </message>
            """
            )
            on_set_subject.assert_not_called()
            on_set_thread_subject.assert_awaited_once()
            assert on_set_thread_subject.call_args[0] == ("thread-id", "some-subject")

        with unittest.mock.patch("slidge.group.room.LegacyMUC.on_set_subject") as on_set_subject, unittest.mock.patch("slidge.group.room.LegacyMUC.on_set_thread_subject") as on_set_thread_subject:
            self.recv(  # language=XML
                """
            <message xmlns="jabber:component:accept"
                     type="groupchat"
                     from="romeo@montague.lit/movim"
                     to="group@aim.shakespeare.lit">
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="uuid"
                         by="group@aim.shakespeare.lit" />
              <thread>thread-id</thread>
              <subject>some-subject</subject>
              <body>some-body</body>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="uuid" />
            </message>
            """
            )
            on_set_subject.assert_not_called()
            on_set_thread_subject.assert_not_called()



import logging
import sys
import traceback
from pathlib import Path

from alembic import command
from alembic.config import Config


def get_alembic_cfg() -> Config:
    alembic_cfg = Config()
    alembic_cfg.set_section_option(
        "alembic",
        "script_location",
        str(Path(__file__).parent / "db" / "alembic"),
    )
    return alembic_cfg


def migrate() -> None:
    try:
        command.upgrade(get_alembic_cfg(), "head")
    except Exception as e:
        traceback.print_exception(e)
        print(
            "Something went wrong during the migration. "
            "This is expected if you upgrade from slidge 0.2, in this case you need to start from a fresh database."
        )
        exit(1)


def main() -> None:
    """
    Updates the (dev) database in ./dev/slidge.sqlite and generates a revision

    Usage: python -m slidge.migration "Revision message blah blah blah"
    """
    dev_db = Path(".") / "dev" / "slidge.sqlite"
    if dev_db.exists():
        # always start from a clean state
        dev_db.unlink()
    alembic_cfg = get_alembic_cfg()
    command.upgrade(alembic_cfg, "head")
    command.revision(alembic_cfg, sys.argv[1], autogenerate=True)


log = logging.getLogger(__name__)

if __name__ == "__main__":
    main()



from conftest import AvatarFixtureMixin

from slidge import BaseGateway, BaseSession
from slidge.contact import LegacyContact, LegacyRoster
from slidge.core.session import _sessions
from slidge.group import LegacyBookmarks
from slidge.util.test import SlidgeTest


class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"


class Session(BaseSession):
    async def login(self):
        await self.contacts.by_name("some id", "some name")
        await self.bookmarks.by_name("some group id", "some group name")
        return "YUP"


class Roster(LegacyRoster):
    async def by_name(self, legacy_id: str, name: str):
        c = await self.by_legacy_id(legacy_id)
        c.name = name
        return c


class Contact(LegacyContact):
    def use_contact_info(self, name: str):
        self.name = name


class Bookmarks(LegacyBookmarks):
    async def fill(self):
        return

    async def by_name(self, legacy_id: str, name: str):
        muc = await self.by_legacy_id(legacy_id)
        muc.name = name


class TestSetNameBeforeFill(AvatarFixtureMixin, SlidgeTest):
    plugin = globals()

    def setUp(self):
        super().setUp()
        self.setup_logged_session(1)

    def test_set_contact_name_before_fill(self):
        contact = self.run_coro(self.romeo.contacts.by_legacy_id("some id"))
        assert contact.name == "some name"
        muc = self.run_coro(self.romeo.bookmarks.by_legacy_id("some group id"))
        assert muc.name == "some group name"



import datetime
import unittest.mock

import sqlalchemy as sa
from slixmpp import Message

from slidge import BaseGateway, BaseSession
from slidge.core.session import _sessions
from slidge.db.models import ArchivedMessage
from slidge.util.archive_msg import HistoryMessage
from slidge.util.test import SlidgeTest


class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"


class Session(BaseSession):
    async def login(self):
        return "YUP"


class TestBackfill(SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        self.setup_logged_session()
        self.xmpp.LEGACY_MSG_ID_TYPE = int

    def tearDown(self):
        with self.xmpp.store.session() as orm:
            orm.execute(sa.delete(ArchivedMessage))
        super().tearDown()

    def test_empty_archive(self):
        with unittest.mock.patch("slidge.group.LegacyMUC.backfill") as backfill:
            self.run_coro(self.room._LegacyMUC__fill_history())
        backfill.assert_awaited_with(None, None)

    def test_live_no_id_before_backfill(self):
        self.first_witch.send_text("BODY 1")
        self.first_witch.send_text("BODY 2")
        self.first_witch.send_text(
            "BODY 3", when=datetime.datetime.now(tz=datetime.timezone.utc)
        )

        with unittest.mock.patch("slidge.group.LegacyMUC.backfill") as backfill:
            self.run_coro(self.room._LegacyMUC__fill_history())
        backfill.assert_awaited_with(None, None)

    def test_live_with_id_before_backfill(self):
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        self.first_witch.send_text("BODY 2", 222, when=now)
        self.first_witch.send_text(
            "BODY 1", 111, when=now - datetime.timedelta(hours=1)
        )
        self.first_witch.send_text(
            "BODY 3", 333, when=now + datetime.timedelta(hours=1)
        )
        with unittest.mock.patch("slidge.group.LegacyMUC.backfill") as backfill:
            self.run_coro(self.room._LegacyMUC__fill_history())
        backfill.assert_awaited_once()
        after, before = backfill.call_args[0]
        assert before.id == 111
        assert before.timestamp == now - datetime.timedelta(hours=1)
        assert after is None

    def _add_back_filled_msg(self, legacy_id=None, when=None):
        with self.xmpp.store.session() as orm:
            self.xmpp.store.mam.add_message(
                orm,
                self.room.stored.id,
                HistoryMessage(Message(), when),
                archive_only=True,
                legacy_msg_id=legacy_id,
            )
            orm.commit()

    def test_pre_backfilled_no_id(self):
        self._add_back_filled_msg()
        with unittest.mock.patch("slidge.group.LegacyMUC.backfill") as backfill:
            self.run_coro(self.room._LegacyMUC__fill_history())
        backfill.assert_awaited_with(None, None)

    def test_pre_backfilled_with_id(self):
        self._add_back_filled_msg(None)
        self._add_back_filled_msg(111)
        self._add_back_filled_msg(222)
        self._add_back_filled_msg(None)
        with unittest.mock.patch("slidge.group.LegacyMUC.backfill") as backfill:
            self.run_coro(self.room._LegacyMUC__fill_history())
        backfill.assert_awaited_once()
        after, before = backfill.call_args[0]
        assert before is None
        assert after is not None
        assert after.id == 222

    def test_pre_backfilled_with_id_and_live(self):
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        self._add_back_filled_msg(None, now - datetime.timedelta(days=5))
        self._add_back_filled_msg(111, now - datetime.timedelta(days=4))
        self._add_back_filled_msg(222, now - datetime.timedelta(days=3))
        self._add_back_filled_msg(None, now - datetime.timedelta(days=2))

        self.first_witch.send_text("BODY1", None)
        self.first_witch.send_text("BODY2", 555)
        self.first_witch.send_text("BODY3", None)
        self.first_witch.send_text("BODY5", 666)
        self.first_witch.send_text("BODY6", None)
        self.first_witch.send_text("BODY7", None)

        with unittest.mock.patch("slidge.group.LegacyMUC.backfill") as backfill:
            self.run_coro(self.room._LegacyMUC__fill_history())
        backfill.assert_awaited_once()
        after, before = backfill.call_args[0]
        assert before.id == 555
        assert after.id == 222



"""Add unique constraint for occupant ID

Revision ID: 805fdc0bf984
Revises: b7a06a86416a
Create Date: 2025-12-17 19:55:11.058263

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "805fdc0bf984"
down_revision: str | None = "b7a06a86416a"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    with op.batch_alter_table("participant", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "uq_participant_room_id_occupant_id", ["room_id", "occupant_id"]
        )


def downgrade() -> None:
    with op.batch_alter_table("participant", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("uq_participant_room_id_occupant_id"), type_="unique"
        )



import asyncio
import os
import shutil
import tempfile
import unittest
from contextlib import asynccontextmanager
from datetime import datetime
from pathlib import Path
from unittest.mock import ANY, MagicMock, patch

import pytest
from slixmpp.plugins.xep_0356.permissions import MessagePermission

from conftest import AvatarFixtureMixin
from test_shakespeare import Base as Shakespeare

from slidge.core import config
from slidge.util.types import LegacyAttachment, MessageReference


@pytest.fixture(scope="function")
def attachment(request, avatar_path):
    class MockHeadResponse:
        status = 200

    class MockGetResponse:
        @staticmethod
        def raise_for_status():
            pass

        @staticmethod
        async def read():
            return avatar_path.read_bytes()

    class MockAioHTTP:
        @asynccontextmanager
        async def head(*a, **k):
            yield MockHeadResponse

        @asynccontextmanager
        async def get(url, *a, **k):
            if "fail" in url:
                raise RuntimeError("IT FAILED")
            yield MockGetResponse

    with (
        patch(
            "slidge.core.mixins.attachment.AttachmentMixin._AttachmentMixin__upload",
            return_value="http://url",
        ) as http_upload,
        patch("aiohttp.ClientSession", return_value=MockAioHTTP) as client_session,
        patch("slidge.core.mixins.attachment.uuid4", return_value="uuid"),
    ):
        request.cls.head = client_session.head = MockAioHTTP.head
        request.cls.http_upload = http_upload
        yield


@pytest.mark.usefixtures("avatar")
@pytest.mark.usefixtures("attachment")
class Base(Shakespeare, AvatarFixtureMixin):
    http_upload: MagicMock

    def _assert_body(self, text="body", i=None):
        if i:
            self.send(  # language=XML
                f"""
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit"
                     id="{i}">
              <body>{text}</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <store xmlns="urn:xmpp:hints" />
            </message>
            """,
                use_values=False,
            )
        else:
            self.send(  # language=XML
                f"""
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <body>{text}</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <store xmlns="urn:xmpp:hints" />
            </message>
            """,
                use_values=False,
            )

    def _assert_file(self, url="http://url", disposition: str = '', local_path: Path | None = None):
        when = (
            datetime.fromtimestamp((local_path or self.avatar_path).stat().st_mtime)
            .isoformat()
            .replace("+00:00", "Z")
        )
        if disposition:
            el = f"disposition='{disposition}'"
        else:
            el = ""
        self.send(  # language=XML
            f"""<message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <reference xmlns="urn:xmpp:reference:0"
                         type="data">
                <media-sharing xmlns="urn:xmpp:sims:1">
                  <sources>
                    <reference xmlns="urn:xmpp:reference:0"
                               uri="{url}"
                               type="data" />
                  </sources>
                  <file xmlns="urn:xmpp:jingle:apps:file-transfer:5">
                    <media-type>image/png</media-type>
                    <name>5x5.png</name>
                    <size>547</size>
                    <date>{when}</date>
                    <hash xmlns="urn:xmpp:hashes:2"
                          algo="sha-256">NdpqDQuHlshve2c0iU25l2KI4cjpoyzaTk3a/CdbjPQ=</hash>
                    <thumbnail xmlns="urn:xmpp:thumbs:1"
                               width="5"
                               height="5"
                               media-type="image/thumbhash"
                               uri="data:image/thumbhash;base64,AAgCBwAAAAAAAAAAAAAAAAAAAAAAAAAA" />
                  </file>
                </media-sharing>
              </reference>
              <file-sharing xmlns="urn:xmpp:sfs:0" {el}>
                <sources>
                  <url-data xmlns="http://jabber.org/protocol/url-data"
                            target="{url}" />
                </sources>
                <file xmlns="urn:xmpp:file:metadata:0">
                  <media-type>image/png</media-type>
                  <name>5x5.png</name>
                  <size>547</size>
                  <date>{when}</date>
                  <hash xmlns="urn:xmpp:hashes:2"
                        algo="sha-256">NdpqDQuHlshve2c0iU25l2KI4cjpoyzaTk3a/CdbjPQ=</hash>
                  <thumbnail xmlns="urn:xmpp:thumbs:1"
                             width="5"
                             height="5"
                             media-type="image/thumbhash"
                             uri="data:image/thumbhash;base64,AAgCBwAAAAAAAAAAAAAAAAAAAAAAAAAA" />
                </file>
              </file-sharing>
              <x xmlns="jabber:x:oob">
                <url>{url}</url>
              </x>
              <body>{url}</body>
              <fallback xmlns="urn:xmpp:fallback:0" for="urn:xmpp:sfs:0">
                <body/>
              </fallback>
            </message>
            """,
            use_values=False,
        )


class TestBodyOnly(Base):
    def test_no_file_no_body(self):
        self.run_coro(self.juliet.send_files([]))
        assert self.next_sent() is None

    def test_just_body(self):
        self.run_coro(self.juliet.send_files([], body="body"))
        self._assert_body()
        self.run_coro(self.juliet.send_files([], body="body", body_first=True))
        self._assert_body()
        self.run_coro(self.juliet.send_files([], body="body", legacy_msg_id=12))
        self._assert_body(i=12)


class TestAttachmentUpload(Base):
    def __test_basic(
        self, attachment: LegacyAttachment, upload_args: tuple, disposition: str = ""
    ):
        """
        Basic test that file is uploaded.
        """
        self.run_coro(self.juliet.send_files([attachment]))
        self.http_upload.assert_called_with(*upload_args)
        self._assert_file(disposition=disposition)

    def _test_reuse(self, attachment: LegacyAttachment, upload_args: tuple):
        """
        Basic test the no new file is uploaded when the same attachment is used
        twice.
        """
        self.run_coro(self.juliet.send_files([attachment]))
        self.http_upload.assert_called_with(*upload_args)
        self._assert_file()
        self.http_upload.reset_mock()
        self.run_coro(self.juliet.send_files([attachment]))
        self.http_upload.assert_not_called()
        self._assert_file()

    def test_path(self):
        self.__test_basic(
            LegacyAttachment(path=self.avatar_path),
            (self.avatar_path, None, "image/png"),
        )

    def test_path_attachment(self):
        self.__test_basic(
            LegacyAttachment(path=self.avatar_path, disposition="attachment"),
            (self.avatar_path, None, "image/png"),
            "attachment",
        )

    def test_thumbhash(self):
        self.__test_basic(
            LegacyAttachment(path=self.avatar_path, content_type="image/png"),
            (self.avatar_path, None, "image/png"),
        )

    def test_path_and_id(self):
        self._test_reuse(
            LegacyAttachment(path=self.avatar_path, legacy_file_id=1235),
            (ANY, None, "image/png"),
        )

    def test_bytes(self):
        with patch("pathlib.Path.stat", return_value=os.stat(self.avatar_path)):
            self.__test_basic(
                LegacyAttachment(data=self.avatar_path.read_bytes(), name="5x5.png"),
                (ANY, "5x5.png", "image/png"),
            )

    def test_bytes_and_id(self):
        with patch("pathlib.Path.stat", return_value=os.stat(self.avatar_path)):
            self._test_reuse(
                LegacyAttachment(
                    data=self.avatar_path.read_bytes(),
                    legacy_file_id=123,
                    name="5x5.png",
                ),
                (ANY, "5x5.png", "image/png"),
            )

    def test_race_condition(self):
        async def stream(t):
            await asyncio.sleep(t)
            for _ in range(10):
                yield b"e"

        self.run_coro(
            asyncio.gather(
                self.juliet.send_file(LegacyAttachment(aio_stream=stream(0.1), legacy_file_id=123)),
                self.juliet.send_file(LegacyAttachment(aio_stream=stream(0.5), legacy_file_id=123)),
            )
        )
        _msg1 = self.next_sent()
        _msg2 = self.next_sent()
        assert self.next_sent() is None

    def test_fail(self):
        self.run_coro(self.juliet.send_files([LegacyAttachment(name="chaton.jpg", url="fail", caption="trop mimi")]))
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <body>/me tried to send a file (chaton.jpg: trop mimi), but something went wrong: IT FAILED.</body>
            </message>
            """
        )


class TestAttachmentNoUpload(Base):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        config.NO_UPLOAD_URL_PREFIX = "https://url"

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        config.NO_UPLOAD_PATH = None
        config.NO_UPLOAD_URL_PREFIX = None

    def setUp(self):
        super().setUp()
        config.NO_UPLOAD_PATH = tempfile.TemporaryDirectory().name

    def tearDown(self):
        super().tearDown()
        try:
            shutil.rmtree(config.NO_UPLOAD_PATH)
        except FileNotFoundError:
            pass

    def __test_basic(self, attachment: LegacyAttachment, url: str, local_path: Path | None = None):
        """
        Basic test that file is copied.
        """
        self.run_coro(self.juliet.send_files([attachment]))
        self._assert_file(url=url, local_path=local_path)

    def __test_reuse(self, attachment: LegacyAttachment, url: str):
        """
        Basic test the no new file is copied when the same attachment is used
        twice.
        """
        self.run_coro(self.juliet.send_files([attachment]))
        self._assert_file(url=url)
        self.run_coro(self.juliet.send_files([attachment]))
        self._assert_file(url=url)

    def test_path(self):
        self.__test_basic(
            LegacyAttachment(path=self.avatar_path), "https://url/uuid/uuid/5x5.png"
        )

    def test_path_and_id(self):
        self.__test_reuse(
            LegacyAttachment(path=self.avatar_path, legacy_file_id=1234),
            "https://url/1234/uuid/5x5.png",
        )

    def test_multi(self):
        self.xmpp.LEGACY_MSG_ID_TYPE = int
        self.xmpp.use_message_ids = True
        self.run_coro(
            self.juliet.send_files(
                [
                    LegacyAttachment(path=self.avatar_path),
                    LegacyAttachment(path=self.avatar_path, caption="CAPTION"),
                ],
                legacy_msg_id=6666,
                body="BODY",
            )
        )
        xmpp_ids = []
        for _ in range(2):
            att = self.next_sent()
            xmpp_ids.append(att.get_id())
        caption = self.next_sent()
        assert caption["body"] == "CAPTION"
        xmpp_ids.append(caption.get_id())
        body = self.next_sent()
        assert body["body"] == "BODY"
        xmpp_ids.append(body.get_id())
        assert self.next_sent() is None
        assert len(set(xmpp_ids)) == len(xmpp_ids)
        self.juliet.react(6666, "♥")
        for _ in range(len(xmpp_ids)):
            reaction = self.next_sent()
            assert reaction["reactions"]["id"] in xmpp_ids

        self.recv(  # language=XML
            """
            <message to="aim.shakespeare.lit"
                     from="montague.lit">
              <privilege xmlns="urn:xmpp:privilege:2">
                <perm access="roster"
                      type="both" />
                <perm access="message"
                      type="outgoing" />
              </privilege>
            </message>
            """
        )
        for i in xmpp_ids:
            with patch("test_shakespeare.Session.on_react") as mock:
                self.recv(  # language=XML
                    f"""
            <message from="romeo@montague.lit/gajim"
                     to="juliet@{self.xmpp.boundjid.bare}/slidge">
              <reactions id='{i}'
                         xmlns='urn:xmpp:reactions:0'>
                <reaction>👋</reaction>
                <reaction>🐢</reaction>
              </reactions>
            </message>
            """
                )
            for j in [k for k in xmpp_ids if k != i]:
                reac = self.next_sent()
                assert reac["privilege"]["forwarded"]["message"]["reactions"]["id"] == j

            mock.assert_awaited_once()
            assert mock.call_args[0][0].jid == self.juliet.jid
            assert mock.call_args[0][1] == 6666
            assert mock.call_args[0][2] == ["👋", "🐢"]
            assert mock.call_args[1] == dict(thread=None)
        self.xmpp.use_message_ids = False
        self.xmpp.LEGACY_MSG_ID_TYPE = True

    def test_multiatt_mark_all_messages_body_caption(self):
        self.xmpp.MARK_ALL_MESSAGES = True
        self.run_coro(
            self.juliet.send_files(
                [
                    LegacyAttachment(path=self.avatar_path),
                ],
                legacy_msg_id="legacy-real-id",
            )
        )
        assert self.next_sent() is not None
        self.run_coro(
            self.juliet.send_files(
                [
                    LegacyAttachment(path=self.avatar_path),
                ],
                legacy_msg_id="legacy-real-id",
            )
        )
        assert self.next_sent() is not None
        self.xmpp.MARK_ALL_MESSAGES = False


    def test_multi_from_user(self):
        self.xmpp["xep_0356"].granted_privileges["montague.lit"].message = MessagePermission.OUTGOING
        self.run_coro(self.juliet.send_files([], body="test", legacy_msg_id="test-msg-id", carbon=True))
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     to="montague.lit"
                     from="aim.shakespeare.lit">
              <privilege xmlns="urn:xmpp:privilege:2">
                <forwarded xmlns="urn:xmpp:forward:0">
                  <message xmlns="jabber:client"
                           type="chat"
                           from="romeo@montague.lit"
                           id="test-msg-id"
                           to="juliet@aim.shakespeare.lit">
                    <body>test</body>
                    <active xmlns="http://jabber.org/protocol/chatstates" />
                    <store xmlns="urn:xmpp:hints" />
                    <markable xmlns="urn:xmpp:chat-markers:0" />
                  </message>
                </forwarded>
              </privilege>
            </message>
            """,
            # use_values=False,
        )

        self.juliet.react(legacy_msg_id="test-msg-id", emojis=["🏥"])
        self.send(  # language=XML
            """
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <store xmlns="urn:xmpp:hints" />
              <reactions xmlns="urn:xmpp:reactions:0"
                         id="test-msg-id">
                <reaction>🏥</reaction>
              </reactions>
            </message>
            """,
            use_values=False,
        )


    def test_multi_from_user_muc(self):
        session = self.get_romeo_session()
        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
        muc.add_user_resource("gajim")
        user_part = self.run_coro(muc.get_user_participant())
        self.run_coro(user_part.send_files([], body="test", legacy_msg_id="test-msg-id"))
        self.send(  # language=XML
            """
            <message type="groupchat"
                     from="room@aim.shakespeare.lit/romeo"
                     id="test-msg-id"
                     to="romeo@montague.lit/gajim">
              <body>test</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="test-msg-id"
                         by="room@aim.shakespeare.lit" />
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="slidge-user" />
            </message>
            """,
            use_values=False,
        )
        with unittest.mock.patch("uuid.uuid4") as uuid:
            uuid.return_value = "prout"
            user_part.react(legacy_msg_id="test-msg-id", emojis=["🏥"])
            self.send(  # language=XML
                """
            <message type="groupchat"
                     from="room@aim.shakespeare.lit/romeo"
                     to="romeo@montague.lit/gajim">
              <store xmlns="urn:xmpp:hints" />
              <stanza-id xmlns="urn:xmpp:sid:0"
                         id="prout"
                         by="room@aim.shakespeare.lit" />
              <reactions xmlns="urn:xmpp:reactions:0"
                         id="test-msg-id">
                <reaction>🏥</reaction>
              </reactions>
              <occupant-id xmlns="urn:xmpp:occupant-id:0"
                           id="slidge-user" />
            </message>
            """,
                use_values=False,
            )

    def test_multi_moderation(self):
        session = self.get_romeo_session()
        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
        muc.add_user_resource("gajim")
        part = muc.get_system_participant()
        self.run_coro(
            part.send_files(
                [
                    LegacyAttachment(path=self.avatar_path),
                    LegacyAttachment(path=self.avatar_path, caption="CAPTION"),
                ],
                legacy_msg_id="the-real-msg-id",
                body="BODY",
            )
        )
        stanza_ids = []
        while (stanza := self.next_sent()) is not None:
            stanza_ids.append(stanza["stanza_id"]["id"])
        assert len(stanza_ids) == 4  # 2 attachments, the caption and the body
        assert "the-real-msg-id" in stanza_ids

        part.moderate("the-real-msg-id")

        moderated_ids = []
        while (stanza := self.next_sent()) is not None:
            moderated_ids.append(stanza["retract"]["id"])
        assert set(stanza_ids) == set(moderated_ids)

    def test_multi_retraction(self):
        session = self.get_romeo_session()
        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
        muc.add_user_resource("gajim")
        part = muc.get_system_participant()
        self.run_coro(
            part.send_files(
                [
                    LegacyAttachment(path=self.avatar_path),
                    LegacyAttachment(path=self.avatar_path, caption="CAPTION"),
                ],
                legacy_msg_id="the-real-msg-id",
                body="BODY",
            )
        )
        stanza_ids = []
        while (stanza := self.next_sent()) is not None:
            stanza_ids.append(stanza["stanza_id"]["id"])
        assert len(stanza_ids) == 4  # 2 attachments, the caption and the body
        assert "the-real-msg-id" in stanza_ids

        part.retract("the-real-msg-id")

        retracted_ids = []
        while (stanza := self.next_sent()) is not None:
            retracted_ids.append(stanza["retract"]["id"])
        assert set(stanza_ids) == set(retracted_ids)

    def test_multi_reply(self):
        session = self.get_romeo_session()
        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
        muc.add_user_resource("gajim")
        part = muc.get_system_participant()
        self.run_coro(
            part.send_files(
                [
                    LegacyAttachment(path=self.avatar_path),
                    LegacyAttachment(path=self.avatar_path, caption="CAPTION"),
                ],
                legacy_msg_id=42,
                body="BODY",
            )
        )
        stanza_ids = []
        while (stanza := self.next_sent()) is not None:
            stanza_ids.append(stanza["stanza_id"]["id"])
        assert len(stanza_ids) == 4  # 2 attachments, the caption and the body

        for stanza_id in stanza_ids:
            with unittest.mock.patch("test_shakespeare.Session.on_text") as on_text:
                self.recv(  # language=XML
                    f"""
            <message from='{session.user_jid}/gajim'
                     to='{muc.jid}'
                     type='groupchat'>
              <body>Great idea!</body>
              <reply to='{part.jid}'
                     id='{stanza_id}'
                     xmlns='urn:xmpp:reply:0' />
            </message>
            """
                )
                on_text.assert_awaited_once()
                assert on_text.call_args[1].get("reply_to_msg_id") == 42

    def test_multi_react(self):
        session = self.get_romeo_session()
        muc = self.run_coro(session.bookmarks.by_legacy_id("room"))
        muc.add_user_resource("gajim")
        part = muc.get_system_participant()
        self.run_coro(
            part.send_files(
                [
                    LegacyAttachment(path=self.avatar_path),
                    LegacyAttachment(path=self.avatar_path, caption="CAPTION"),
                ],
                legacy_msg_id=42,
                body="BODY",
            )
        )
        stanza_ids = []
        while (stanza := self.next_sent()) is not None:
            stanza_ids.append(stanza["stanza_id"]["id"])

        part.react(legacy_msg_id=42, emojis=["♥"])

        reacted_stanzas_ids = []
        while (stanza := self.next_sent()) is not None:
            reacted_stanzas_ids.append(stanza["reactions"]["id"])

        assert set(stanza_ids) == set(reacted_stanzas_ids)

    def test_reply_with_attachment(self):
        self.run_coro(
            self.juliet.send_files(
                [
                    LegacyAttachment(path=self.avatar_path),
                ],
                reply_to=MessageReference("some_msg_id", body="a body"),
            )
        )
        when = (
            datetime.fromtimestamp(self.avatar_path.stat().st_mtime)
            .isoformat()
            .replace("+00:00", "Z")
        )
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <reply xmlns="urn:xmpp:reply:0"
                     id="some_msg_id" />
              <body>https://url/uuid/uuid/5x5.png</body>
              <reference xmlns="urn:xmpp:reference:0"
                         type="data">
                <media-sharing xmlns="urn:xmpp:sims:1">
                  <sources>
                    <reference xmlns="urn:xmpp:reference:0"
                               uri="https://url/uuid/uuid/5x5.png"
                               type="data" />
                  </sources>
                  <file xmlns="urn:xmpp:jingle:apps:file-transfer:5">
                    <media-type>image/png</media-type>
                    <name>5x5.png</name>
                    <size>547</size>
                    <date>{when}</date>
                    <hash xmlns="urn:xmpp:hashes:2"
                          algo="sha-256">NdpqDQuHlshve2c0iU25l2KI4cjpoyzaTk3a/CdbjPQ=</hash>
                    <thumbnail xmlns="urn:xmpp:thumbs:1"
                               width="5"
                               height="5"
                               media-type="image/thumbhash"
                               uri="data:image/thumbhash;base64,AAgCBwAAAAAAAAAAAAAAAAAAAAAAAAAA" />
                  </file>
                </media-sharing>
              </reference>
              <file-sharing xmlns="urn:xmpp:sfs:0">
                <sources>
                  <url-data xmlns="http://jabber.org/protocol/url-data"
                            target="https://url/uuid/uuid/5x5.png" />
                </sources>
                <file xmlns="urn:xmpp:file:metadata:0">
                  <media-type>image/png</media-type>
                  <name>5x5.png</name>
                  <size>547</size>
                  <date>{when}</date>
                  <hash xmlns="urn:xmpp:hashes:2"
                        algo="sha-256">NdpqDQuHlshve2c0iU25l2KI4cjpoyzaTk3a/CdbjPQ=</hash>
                  <thumbnail xmlns="urn:xmpp:thumbs:1"
                             width="5"
                             height="5"
                             media-type="image/thumbhash"
                             uri="data:image/thumbhash;base64,AAgCBwAAAAAAAAAAAAAAAAAAAAAAAAAA" />
                </file>
              </file-sharing>
              <x xmlns="jabber:x:oob">
                <url>https://url/uuid/uuid/5x5.png</url>
              </x>
              <fallback xmlns="urn:xmpp:fallback:0"
                        for="urn:xmpp:sfs:0">
                <body />
              </fallback>
            </message>
            """,
            use_values=False,
        )

    def test_url(self):
        tmp = tempfile.mkdtemp()

        with unittest.mock.patch("tempfile.mkdtemp", return_value=tmp):
            self.__test_basic(
                LegacyAttachment(url=self.avatar_url, name="5x5.png"),
                url="https://url/uuid/uuid/5x5.png",
                local_path=Path(tmp) / "5x5.png",
            )
        shutil.rmtree(tmp)


class TestAttachmentOriginalUrl(Base):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        config.USE_ATTACHMENT_ORIGINAL_URLS = True


    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        config.USE_ATTACHMENT_ORIGINAL_URLS = False

    def __check(self, url, name=None, content_type=None, disposition=None):
        disposition = f"disposition='{disposition}'" if disposition else ""
        # file_sharing = "<file-sharing xmlns='urn:xmpp:sfs:0'" + disposition
        name = f"<name>{name}</name>" if name else ""
        content_type = f"<media-type>{content_type}</media-type>" if content_type else ""

        self.send(  # language=XML
            f"""<message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <media-sharing xmlns="urn:xmpp:sims:1">
                <sources>
                  <reference xmlns="urn:xmpp:reference:0"
                             uri="{url}"
                             type="data" />
                </sources>
                <file xmlns="urn:xmpp:jingle:apps:file-transfer:5">{content_type} {name}</file>
              </media-sharing>
              <file-sharing xmlns="urn:xmpp:sfs:0" {disposition}>
                <sources>
                  <url-data xmlns="http://jabber.org/protocol/url-data"
                            target="{url}" />
                </sources>
                <file xmlns="urn:xmpp:file:metadata:0">{content_type} {name}</file>
              </file-sharing>
              <x xmlns="jabber:x:oob">
                <url>{url}</url>
              </x>
              <fallback xmlns='urn:xmpp:fallback:0'
                        for='urn:xmpp:sfs:0'>
                <body />
              </fallback>
              <body>{url}</body>
            </message>
            """
        )

    def __check_no_meta(self, url):
        self.send(  # language=XML
            f"""
            <message type="chat"
                     from="juliet@aim.shakespeare.lit/slidge"
                     to="romeo@montague.lit">
              <x xmlns="jabber:x:oob">
                <url>{url}</url>
              </x>
              <body>{url}</body>
            </message>
            """
        )

    def __send_file(self, url, name=None, content_type=None, disposition=None):
        self.run_coro(self.juliet.send_file(
            LegacyAttachment(url=url, name=name, content_type=content_type, disposition=disposition)))

    def test_url_original(self):
        self.__send_file("prout", name="5x5.png", disposition="inline")
        self.__check("prout", name="5x5.png", content_type="image/png", disposition="inline")

        self.__send_file("prout", name="5x5.png", content_type="GNAGNA")
        self.__check("prout", name="5x5.png", content_type="GNAGNA")

        self.__send_file("prout", disposition="inline")
        self.__check("prout", disposition="inline")

        self.__send_file("prout", name="5x5.png")
        self.__check("prout", name="5x5.png", content_type="image/png")

        self.__send_file("prout.png")
        self.__check("prout.png", content_type="image/png")

    def test_no_metadata(self):
        self.__send_file("prout")
        self.__check_no_meta("prout")



from slidge import BaseGateway, BaseSession, LegacyContact, LegacyMUC, LegacyBookmarks, LegacyRoster, LegacyParticipant
from slidge.util.test import SlidgeTest
from slixmpp.exceptions import XMPPError

class Gateway(BaseGateway):
    COMPONENT_NAME = "Megacorp Chat XMPP Gateway"
    GROUPS = True


class Session(BaseSession):
    async def login(self):
        pass


class Contact(LegacyContact):
    pass


class Roster(LegacyRoster):
    async def legacy_id_to_jid_username(self, legacy_id: str) -> str:
        if not legacy_id.startswith("room-"):
            raise XMPPError


class Bookmarks(LegacyBookmarks):
    async def fill(self):
        pass


class TestSpaces(SlidgeTest):
    plugin = globals()

    def setUp(self):
        super().setUp()
        self._add_slidge_user()

    def test_archive_correction_no_event_id(self) -> None:
        muc = self.get_joined_muc("room-correction")
        part: LegacyParticipant = self.run_coro(muc.get_participant("part"))
        part.send_text("a body", "a-legacy-id")
        self.next_sent()  # part presence

        msg = self.next_sent()
        assert msg["body"] == "a body", msg
        assert msg["stanza_id"]["id"] == "a-legacy-id", msg
        part.correct("a-legacy-id", "a new body")
        msg = self.next_sent()
        assert msg["body"] == "a new body", msg
        correction_stanza_id = msg["stanza_id"]["id"]

        messages = list(muc.archive.get_all(before_id=correction_stanza_id))
        assert len(messages) == 1
        assert messages[0]

    def test_archive_correction_specify_event_id(self) -> None:
        muc = self.get_joined_muc("room-correction")
        part: LegacyParticipant = self.run_coro(muc.get_participant("part"))
        part.send_text("a body", "a-legacy-id")
        self.next_sent()  # part presence

        msg = self.next_sent()
        assert msg["body"] == "a body", msg
        assert msg["stanza_id"]["id"] == "a-legacy-id", msg
        part.correct("a-legacy-id", "a new body", correction_event_id="another-legacy-id")
        msg = self.next_sent()
        assert msg["body"] == "a new body", msg
        correction_stanza_id = msg["stanza_id"]["id"]
        assert correction_stanza_id == "another-legacy-id"

        messages = list(muc.archive.get_all(before_id=correction_stanza_id))
        assert len(messages) == 1
        assert messages[0]

        messages = list(muc.get_archived_messages(correction_stanza_id))
        assert len(messages) == 1

    def test_archive_correction_specify_existing_event_id(self) -> None:
        muc = self.get_joined_muc("room-correction")
        part: LegacyParticipant = self.run_coro(muc.get_participant("part"))
        part.send_text("a body", "a-legacy-id")
        self.next_sent()  # part presence

        msg = self.next_sent()
        assert msg["body"] == "a body", msg
        assert msg["stanza_id"]["id"] == "a-legacy-id", msg
        part.correct("a-legacy-id", "a new body", correction_event_id="a-legacy-id")
        msg = self.next_sent()
        assert msg["body"] == "a new body", msg
        correction_stanza_id = msg["stanza_id"]["id"]
        assert correction_stanza_id == "a-legacy-id"

        messages = list(muc.archive.get_all(before_id=correction_stanza_id))
        assert not messages



# Handle slidge commands by exchanging chat messages with the gateway components.

# Ad-hoc methods should provide a better UX, but some clients do not support them,
# so this is mostly a fallback.
import asyncio
import functools
import inspect
import logging
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Literal, overload
from urllib.parse import quote as url_quote

from slixmpp import JID, CoroutineCallback, Message, StanzaPath
from slixmpp.exceptions import XMPPError
from slixmpp.types import JidStr, MessageTypes

from . import Command, CommandResponseType, Confirmation, Form, TableResult
from .categories import CommandCategory

if TYPE_CHECKING:
    from ..core.gateway import BaseGateway


class ChatCommandProvider:
    UNKNOWN = "Wut? I don't know that command: {}"

    def __init__(self, xmpp: "BaseGateway") -> None:
        self.xmpp = xmpp
        self._keywords = list[str]()
        self._commands: dict[str, Command] = {}
        self._input_futures = dict[str, asyncio.Future[str]]()
        self.xmpp.register_handler(
            CoroutineCallback(
                "chat_command_handler",
                StanzaPath(f"message@to={self.xmpp.boundjid.bare}"),
                self._handle_message,  # type: ignore
            )
        )

    def register(self, command: Command):
        """
        Register a command to be used via chat messages with the gateway

        Plugins should not call this, any class subclassing Command should be
        automatically added by slidge core.

        :param command: the new command
        """
        t = command.CHAT_COMMAND
        if t in self._commands:
            raise RuntimeError("There is already a command triggered by '%s'", t)
        self._commands[t] = command

    @overload
    async def input(self, jid: JidStr, text: str | None = None) -> str: ...

    @overload
    async def input(
        self, jid: JidStr, text: str | None = None, *, blocking: Literal[False] = ...
    ) -> asyncio.Future[str]: ...

    @overload
    async def input(
        self,
        jid: JidStr,
        text: str | None = None,
        *,
        mtype: MessageTypes = "chat",
        timeout: int = 60,
        blocking: Literal[True] = True,
        **msg_kwargs: Any,
    ) -> str: ...

    async def input(
        self,
        jid: JidStr,
        text: str | None = None,
        *,
        mtype: MessageTypes = "chat",
        timeout: int = 60,
        blocking: bool = True,
        **msg_kwargs: Any,
    ) -> str | asyncio.Future[str]:
        """
        Request arbitrary user input using a simple chat message, and await the result.

        You shouldn't need to call directly bust instead use :meth:`.BaseSession.input`
        to directly target a user.

        NB: When using this, the next message that the user sent to the component will
        not be transmitted to :meth:`.BaseGateway.on_gateway_message`, but rather intercepted.
        Await the coroutine to get its content.

        :param jid: The JID we want input from
        :param text: A prompt to display for the user
        :param mtype: Message type
        :param timeout:
        :param blocking: If set to False, timeout has no effect and an :class:`asyncio.Future`
            is returned instead of a str
        :return: The user's reply
        """
        jid = JID(jid)
        if text is not None:
            self.xmpp.send_message(
                mto=jid,
                mbody=text,
                mtype=mtype,
                mfrom=self.xmpp.boundjid.bare,
                **msg_kwargs,
            )
        f = asyncio.get_event_loop().create_future()
        self._input_futures[jid.bare] = f
        if not blocking:
            return f
        try:
            await asyncio.wait_for(f, timeout)
        except TimeoutError:
            self.xmpp.send_message(
                mto=jid,
                mbody="You took too much time to reply",
                mtype=mtype,
                mfrom=self.xmpp.boundjid.bare,
            )
            del self._input_futures[jid.bare]
            raise XMPPError("remote-server-timeout", "You took too much time to reply")

        return f.result()

    async def _handle_message(self, msg: Message):
        if not msg["body"]:
            return

        if not msg.get_from().node:
            return  # ignore component and server messages

        f = self._input_futures.pop(msg.get_from().bare, None)
        if f is not None:
            f.set_result(msg["body"])
            return

        c = msg["body"]
        first_word, *rest = c.split(" ")
        first_word = first_word.lower()

        if first_word == "help":
            return self._handle_help(msg, *rest)

        mfrom = msg.get_from()

        command = self._commands.get(first_word)
        if command is None:
            return self._not_found(msg, first_word)

        try:
            session = command.raise_if_not_authorized(mfrom)
        except XMPPError as e:
            reply = msg.reply()
            reply["body"] = e.text
            reply.send()
            raise

        result = await self.__wrap_handler(msg, command.run, session, mfrom, *rest)
        self.xmpp.delivery_receipt.ack(msg)
        return await self._handle_result(result, msg, session)

    def __make_uri(self, body: str) -> str:
        return f"xmpp:{self.xmpp.boundjid.bare}?message;body={body}"

    async def _handle_result(self, result: CommandResponseType, msg: Message, session):
        if isinstance(result, str) or result is None:
            reply = msg.reply()
            reply["body"] = result or "End of command."
            reply.send()
            return

        if isinstance(result, Form):
            try:
                return await self.__handle_form(result, msg, session)
            except XMPPError as e:
                if (
                    result.timeout_handler is None
                    or e.condition != "remote-server-timeout"
                ):
                    raise e
                return result.timeout_handler()

        if isinstance(result, Confirmation):
            yes_or_no = await self.input(msg.get_from(), result.prompt)
            if not yes_or_no.lower().startswith("y"):
                reply = msg.reply()
                reply["body"] = "Canceled"
                reply.send()
                return
            result = await self.__wrap_handler(
                msg,
                result.handler,
                session,
                msg.get_from(),
                *result.handler_args,
                **result.handler_kwargs,
            )
            return await self._handle_result(result, msg, session)

        if isinstance(result, TableResult):
            if len(result.items) == 0:
                msg.reply("Empty results").send()
                return

            body = result.description + "\n"
            for item in result.items:
                for f in result.fields:
                    if f.type == "jid-single":
                        j = JID(item[f.var])
                        value = f"xmpp:{percent_encode(j)}"
                        if result.jids_are_mucs:
                            value += "?join"
                    else:
                        value = item[f.var]  # type:ignore
                    body += f"\n{f.label or f.var}: {value}"
            msg.reply(body).send()

    async def __handle_form(self, result: Form, msg: Message, session):
        form_values = {}
        for t in result.title, result.instructions:
            if t:
                msg.reply(t).send()
        for f in result.fields:
            if f.type == "fixed":
                msg.reply(f"{f.label or f.var}: {f.value}").send()
            else:
                if f.type == "list-multi":
                    msg.reply(
                        "Multiple selection allowed, use new lines as a separator, ie, "
                        "one selected item per line. To select no item, reply with a space "
                        "(the punctuation)."
                    ).send()
                if f.options:
                    for o in f.options:
                        msg.reply(f"{o['label']}: {self.__make_uri(o['value'])}").send()
                if f.value:
                    msg.reply(f"Default: {f.value}").send()
                if f.type == "boolean":
                    msg.reply("yes: " + self.__make_uri("yes")).send()
                    msg.reply("no: " + self.__make_uri("no")).send()

                ans = await self.xmpp.input(
                    msg.get_from(), (f.label or f.var) + "? (or 'abort')"
                )
                if ans.lower() == "abort":
                    return await self._handle_result("Command aborted", msg, session)
                if f.type == "boolean":
                    if ans.lower() == "yes":
                        ans = "true"
                    else:
                        ans = "false"

                if f.type.endswith("multi"):
                    choices = [] if ans == " " else ans.split("\n")
                    form_values[f.var] = f.validate(choices)
                else:
                    form_values[f.var] = f.validate(ans)
        result = await self.__wrap_handler(
            msg,
            result.handler,
            form_values,
            session,
            msg.get_from(),
            *result.handler_args,
            **result.handler_kwargs,
        )
        return await self._handle_result(result, msg, session)

    @staticmethod
    async def __wrap_handler(msg, f: Callable | functools.partial, *a, **k):
        try:
            if inspect.iscoroutinefunction(f):
                return await f(*a, **k)
            elif hasattr(f, "func") and inspect.iscoroutinefunction(f.func):
                return await f(*a, **k)
            else:
                return f(*a, **k)
        except Exception as e:
            log.debug("Error in %s", f, exc_info=e)
            reply = msg.reply()
            reply["body"] = f"Error: {e}"
            reply.send()

    def _handle_help(self, msg: Message, *rest) -> None:
        if len(rest) == 0:
            reply = msg.reply()
            reply["body"] = self._help(msg.get_from())
            reply.send()
        elif len(rest) == 1 and (command := self._commands.get(rest[0])):
            reply = msg.reply()
            reply["body"] = f"{command.CHAT_COMMAND}: {command.NAME}\n{command.HELP}"
            reply.send()
        else:
            self._not_found(msg, str(rest))

    def _help(self, mfrom: JID):
        session = self.xmpp.get_session_from_jid(mfrom)

        msg = "Available commands:"
        for c in sorted(
            self._commands.values(),
            key=lambda co: (
                (
                    co.CATEGORY
                    if isinstance(co.CATEGORY, str)
                    else (
                        co.CATEGORY.name
                        if isinstance(co.CATEGORY, CommandCategory)
                        else ""
                    )
                ),
                co.CHAT_COMMAND,
            ),
        ):
            try:
                c.raise_if_not_authorized(mfrom, fetch_session=False, session=session)
            except XMPPError:
                continue
            msg += f"\n{c.CHAT_COMMAND} -- {c.NAME}"
        return msg

    def _not_found(self, msg: Message, word: str):
        e = self.UNKNOWN.format(word)
        msg.reply(e).send()
        raise XMPPError("item-not-found", e)


def percent_encode(jid: JID) -> str:
    return f"{url_quote(jid.user)}@{jid.server}"  # type:ignore


log = logging.getLogger(__name__)
